/*************************************/ 
/* Original oroborus window manager  */
/* (c) Ken Lynch Jan 2001            */
/*                                   */
/* Extra functionality for OroborOSX */
/* (c) Adrian Umpleby Jan 2002       */
/*                                   */
/* Distributed under the GPL         */ 
/* See LICENSE for more details      */ 
/*************************************/ 
 
#include "oroborus.h" 
#include "hackcmds.h"

void add_client_to_group(Client *c)
{
  long i;
  int grouplen;
  Client *ac, *fc=NULL;
  
  sprintf(lastsub,"TRACE2: add_client_to_group  win: %#lx (last: %#lx)\n",C_WINDOW,FC_WINDOW); 
  subtrace();
#ifdef SHOW_TRACE2 
  printf("TRACE2: add_client_to_group  win: %#lx (lats: %#lx)\n",C_WINDOW,FC_WINDOW); 
#endif

  if(c==NULL) return;
  if(client_list==NULL) return;
  if(client_count==0) return;
  if(c->osxwinfo.winref==0) return;
  if(c->clientgroup==NULL) return;
  
  grouplen=strlen(c->clientgroup);
// Loop through the clients, looking for matches with group of latest client
  ac=client_list;
  for(i=0;i<client_count;i++) {
    if(ac->osxwinfo.winref!=0)
      if(ac->clientgroup)
        if(strlen(ac->clientgroup)==grouplen)
          if(!strncmp(ac->clientgroup,c->clientgroup,grouplen)) {
            if(fc==NULL) { // check if this is the first match...
                fc = ac;
                if(fc->osxwinfo.group_winref==0) {
                    // only one other client in this group, so create a group proxy window
                    // (also sets up the client's group_winref)
                    MakeGroupWindow(&(c->osxwinfo),c->clientgroup);
                    // update the other client's group proxy window
                    fc->osxwinfo.group_winref = c->osxwinfo.group_winref;
                    // make sure its title changes appropriately
                    CheckOSXWinTitle(&(fc->osxwinfo),fc->name,fc->clientgroup);
                } else {
                    // set new client's group winref, and make sure group window is reordered
                    c->osxwinfo.group_winref = fc->osxwinfo.group_winref;
                    ReorderGroupWin(fc->osxwinfo.group_winref);
                }
            }
            ReorderOSXWin(&(ac->osxwinfo));
          }
    ac=ac->next;
  }
    
}

void check_delete_group(Client *c)
{
    Client *ac, *lc=NULL;
    long i, cnt=0;
    long group_winref;
    
    if(c==NULL) return;
    group_winref = get_group_for_client(c);
    if(group_winref==0) return;
    if(client_list==NULL) return;
    
    ac = client_list;
    for(i=0;i<client_count;i++) {
        if(get_group_for_client(ac)==group_winref) {
            cnt++;
            if(ac!=c && ac->osxwinfo.group_winref!=0) lc=ac;
        }
        ac = ac->next;
    }

    if(cnt==2 && lc!=NULL) {
        DestroyGroupWindow(lc->osxwinfo.group_winref);
        lc->osxwinfo.group_winref = 0;
        CheckOSXWinTitle(&(lc->osxwinfo),lc->name,lc->clientgroup);
    }
}

long get_group_for_client(Client *c)
{
    long winref=0;
    
    if(c==NULL) return 0;
    
    while(1) {
        winref = c->osxwinfo.group_winref;
        if(winref==0) winref = c->osxwinfo.winref;
        if(winref==0) {
            if(c->parent==NULL) return 0;
            c = c->parent;
        } else
            return winref;
    }
}

void raise_group(long group_winref)
{
    Window *windows=NULL, w1, w2; 
    int count, i;
    Client *c, *lc=NULL, *tc=NULL, *fc=NULL;
    XWindowAttributes attr; 
    
    if(client_count==0) return;
    if(client_list==NULL) return;
    
    if(gVerboseConsole)
        printf("%s info: raising group with winref %#lx\n",PROGNAME,group_winref);
    if(group_winref==0) return;
        
    XQueryTree(dpy, root, &w1, &w2, &windows, &count); 
    for(i=count-1;i>=0;i--) {
        XGetWindowAttributes(dpy, windows[i], &attr); 
        if(attr.map_state==IsViewable) 
        { 
            c=get_client_from_window(windows[i], FRAME); 
            if(c!=NULL && (get_group_for_client(c)==group_winref)) 
            { 
#ifdef LAYER_SUPPORT 
                if(get_gnome_hint(c->window, WIN_LAYER)>=layer) {
#endif
                    if((c->is_docked==0) && (c->is_icon==NOT_ICONBAR)) {
                        if(lc==NULL) {
                            force_raise_client(c);
                            fc = c;
                        } else
                            stack_client_re_client(c,lc,Below);
                        lc = c;
                    }
                    tc = c;
#ifdef LAYER_SUPPORT 
                }
#endif /* LAYER */ 
            } 
        } 
    }
    
    if(windows) XFree(windows);

    if(fc==NULL) fc = tc;
        
    if(fc!=NULL) {
        if(fc->is_icon==NOT_ICONBAR) {
            raise_client(fc);
            set_focus_client(fc);
        } else
            minmax_client(fc);
    }
}

void lower_group(long group_winref)
{
    Window *windows=NULL, w1, w2; 
    int count, i;
    Client *c, *prevc=NULL;
    XWindowAttributes attr; 
    
    if(client_count<=1) return;
    if(client_list==NULL) return;
    
    if(gVerboseConsole)
        printf("%s info: lowering group with winref %#lx\n",PROGNAME,group_winref);
    if(group_winref==0) return;
        
    XQueryTree(dpy, root, &w1, &w2, &windows, &count); 
    for(i=0;i<count;i++) {
        XGetWindowAttributes(dpy, windows[i], &attr); 
        if(attr.map_state==IsViewable) 
        { 
            c=get_client_from_window(windows[i], FRAME); 
            if(c!=NULL && (get_group_for_client(c)==group_winref)) 
            { 
#ifdef LAYER_SUPPORT 
                if(get_gnome_hint(c->window, WIN_LAYER)>=layer) {
#endif
                    if((c->is_docked==0) && (c->is_icon==NOT_ICONBAR)) {
                        if(prevc==NULL)
                            lower_client(c);
                        else
                            stack_client_re_client(c,prevc,Above);
                        prevc = c;
                    }
#ifdef LAYER_SUPPORT 
                }
#endif /* LAYER */ 
            } 
        } 
    }
    
    if(windows) XFree(windows);
}

Client *get_next_group_client() {
    Window *windows=NULL, w1, w2; 
    int count, i;
    Client *c, *bc;
    XWindowAttributes attr; 
    long winref;
    
    if(client_count<=1) return NULL;
    if(client_list==NULL) return NULL;
    
    c = get_top_most_client(WIN_LAYER_NORMAL,0);
    if(c==NULL) return NULL;
    
    winref = get_group_for_client(c);
    if(winref==0) {
        bc = get_bottom_most_client(WIN_LAYER_NORMAL,0);
        if(bc==c) bc = NULL;
        return bc;
    }
    
    XQueryTree(dpy, root, &w1, &w2, &windows, &count); 
    for(i=0;i<count;i++) {
        XGetWindowAttributes(dpy, windows[i], &attr); 
        if(attr.map_state==IsViewable) 
        { 
            c=get_client_from_window(windows[i], FRAME); 
            if(c!=NULL && (get_group_for_client(c)!=winref)) 
            { 
#ifdef LAYER_SUPPORT 
                if(get_gnome_hint(c->window, WIN_LAYER)>=layer) {
#endif
                    if((c->is_docked==0) && (c->is_icon==NOT_ICONBAR)) {
                        if(windows) XFree(windows);
                        if(gVerboseConsole)
                            printf("%s info: got next group with winref %#lx (top: %#lx)\n",PROGNAME,c->osxwinfo.group_winref,winref);
                        return c;
                    }
#ifdef LAYER_SUPPORT 
                }
#endif /* LAYER */ 
            } 
        } 
    }
    
    if(windows) XFree(windows);
    if(gVerboseConsole)
    	printf("%s info: got no next group! (top: %#lx)\n",PROGNAME,winref);
    return NULL;
}

char *get_next_group_name() {
    Client *c;
    char *retval=NULL;
    
    c = get_next_group_client();
    if(c!=NULL) retval = strdup(c->clientgroup);

    return retval;
}

void raise_next_group() {
    Client *c;
    
    c = get_next_group_client();
    if(c!=NULL) {
        if(c->osxwinfo.view_state==VIEWSTATE_HIDDEN)
            map_group(get_group_for_client(c));
        else if(get_group_for_client(c)==0)
            raise_client(c);
        else
            raise_group(get_group_for_client(c));
    }
}

Client *get_next_client_in_group(long group_winref)
{
    Window *windows=NULL, w1, w2; 
    int count, i;
    Client *c=NULL;
    XWindowAttributes attr;
    
    if(client_count<=1) return NULL;
    if(client_list==NULL) return NULL;
    
    if(group_winref==0) {
        c = get_bottom_most_client(WIN_LAYER_NORMAL,0);
        return c;
    } else {
        XQueryTree(dpy, root, &w1, &w2, &windows, &count); 
        for(i=0;i<count;i++) {
            XGetWindowAttributes(dpy, windows[i], &attr); 
            if(attr.map_state==IsViewable) 
            { 
                c=get_client_from_window(windows[i], FRAME); 
                if(c!=NULL && get_group_for_client(c)==group_winref && check_client_modality(c,0)!=get_focus_client()) 
                { 
#ifdef LAYER_SUPPORT 
                    if(get_gnome_hint(c->window, WIN_LAYER)>=layer) {
#endif
                  	 if((c->is_docked==0) && (c->is_icon==NOT_ICONBAR)) {
                            if(gVerboseConsole)
                                printf("%s info: got next in group: %#lx (grpref: %#lx)\n",PROGNAME,c->window,group_winref);
                            if(windows) XFree(windows);
                            return c;
                        }
#ifdef LAYER_SUPPORT 
                    }
#endif /* LAYER */ 
                } 
            } 
        }
    
        if(windows) XFree(windows);
    }
    
    return NULL;
}

char *get_next_title_in_group(int strict)
{
    Client *fc, *c;
    char *wintitle=NULL, *offset_title=NULL;
    
    fc=get_focus_client();
//#ifdef DEBUG
    if(fc==NULL && gVerboseConsole)
      printf("Got NULL focus client in get_next_title_in_group.\n");
//#endif
    if(fc==NULL) return NULL;
    
    if(strict)
        c=get_next_client_in_group(get_group_for_client(fc));
    else
        c=get_next_client_in_group(fc->osxwinfo.group_winref);

    if(c!=NULL) {  
      if(c->name==NULL && c->clientgroup==NULL) {
        wintitle = malloc(15);
        sprintf(wintitle,"?: Untitled");
      } else if(c->name==NULL) {
        wintitle = malloc(strlen(c->clientgroup)+15);
        sprintf(wintitle,"%s: Untitled",c->clientgroup);
      } else if(c->clientgroup==NULL) {
        wintitle = malloc(strlen(c->name)+5);
        sprintf(wintitle,"?: %s",c->name);
      } else {
        offset_title = c->name;
        if(strlen(c->clientgroup)<strlen(c->name))
            if(!strncmp(c->clientgroup,c->name,strlen(c->clientgroup))) {
                offset_title+=strlen(c->clientgroup);
                // strip out spaces after class name
                while(strlen(offset_title)>1 && !strncmp(offset_title," ",1))
                    offset_title++;
                // strip out colon
                if(strlen(offset_title)>1 && !strncmp(offset_title,":",1))
                    offset_title++;
                // strip out any spaces after colon
                while(strlen(offset_title)>1 && !strncmp(offset_title," ",1))
                    offset_title++;
            }
        wintitle = malloc(strlen(c->clientgroup)+strlen(offset_title)+5);
        sprintf(wintitle,"%s:  %s",c->clientgroup,offset_title);
      }
    }
    
    return wintitle;
}

long get_next_osxwinref_in_group(int raiseit, int strict)
{
    Client *c, *fc;

    fc=get_focus_client();
    
//#ifdef DEBUG
    if(fc==NULL && gVerboseConsole)
      printf("Got NULL focus client in get_next_osxwinref_in_group.\n");
//#endif
    if(fc==NULL) return 0;

    if(strict)
        c=get_next_client_in_group(get_group_for_client(fc));
    else
        c=get_next_client_in_group(fc->osxwinfo.group_winref);

    if(c!=NULL) {
      if(raiseit!=0) {
        if(c->is_docked)
          will_undock_client(c);
        else {
          set_focus_client(c);
          if((c->auto_properties & (AUTO_FOCUS|AUTO_RAISE))==AUTO_FOCUS)
            raise_client(c);
          if(c->al_list!=NULL) XFree(c->al_list);
          c->al_list = NULL;
          c->al_count = 0;
        }
      }
      return c->osxwinfo.winref;
    } else
      return 0;
}

Client *get_top_not_this_in_group(Client *notc, int layer, int get_icons, long group_winref)
{
    Window *windows=NULL, w1, w2; 
    int count, i;
    Client *c=NULL;
    XWindowAttributes attr;
    
        XQueryTree(dpy, root, &w1, &w2, &windows, &count); 
        for(i=count-1;i>=0;i--) {
            XGetWindowAttributes(dpy, windows[i], &attr); 
            if(attr.map_state==IsViewable) 
            { 
                c=get_client_from_window(windows[i], FRAME); 
                if(c!=NULL && c!=notc && get_group_for_client(c)==group_winref && check_client_modality(c,0)!=notc) 
                { 
#ifdef LAYER_SUPPORT 
                    if(get_gnome_hint(c->window, WIN_LAYER)>=layer) {
#endif
                  	 if(((c->is_docked==0) && (c->is_icon==NOT_ICONBAR)) || get_icons) {
                            if(gVerboseConsole)
                                printf("%s info: got top in group: %#lx (grpref: %#lx)\n",PROGNAME,c->window,group_winref);
                            if(windows) XFree(windows);
                            return c;
                        }
#ifdef LAYER_SUPPORT 
                    }
#endif /* LAYER */ 
                } 
            } 
        }
    
        if(windows) XFree(windows);
        
    return NULL;
}


void map_group(long group_winref)
{
  Window *windows=NULL, w1, w2; 
  int count, i, x, y; 
  Client *c;
  XWindowAttributes attr;

//  if(can_send_hackcmds!=0) return;
  
    if(client_count==0) return;
    if(client_list==NULL) return;
    
    if(gVerboseConsole)
        printf("%s info: map_group with winref %#lx\n",PROGNAME,group_winref);
    if(group_winref==0) return;

  sprintf(lastsub,"TRACE1: map_group %#lx\n",group_winref); 
  subtrace();
#ifdef SHOW_TRACE1
  printf("TRACE1: map_group %#lx\n",group_winref); 
#endif
 
  MyIncreaseXServerGrab(); // grab the server, so nothing happens while we're messing around
  XQueryTree(dpy, root, &w1, &w2, &windows, &count); 
#ifdef DEBUG
  printf("  children of root: %d\n",count);
#endif

  // Search through the windows that actually exist. (It's possible that some might have
  // disappeared while hidden, but they remain in the client list. This is because the window
  // manager had no notification they were unmapped as they already were unmapped when the
  // user chose to hide OroborOSX.)
  for(i=0;i<count;i++) { 
#ifdef DEBUG
    printf("  child %d (%#lx)\n",i,windows[i]);
#endif
    c=get_client_from_window(windows[i],FRAME);
    if(c!=NULL && get_group_for_client(c)==group_winref) {
      if(can_send_hackcmds==0) {
        if(c->is_docked==0) {
          XMapWindow(dpy, c->frame); 
          if(c->is_icon!=NOT_ICONBAR)
            c->osxwinfo.view_state=VIEWSTATE_ICONBAR;
          else
            c->osxwinfo.view_state=VIEWSTATE_VISIBLE;
          if(c->osxwinfo.winref) ChangedOSXWinState(&(c->osxwinfo),c->name,c->clientgroup);
          if(c->is_icon==NOT_ICONBAR) {
            XMapWindow(dpy, c->window);
            set_wm_state(c->window, NormalState);
          }
          if((c->has_frame) || (c->is_icon!=NOT_ICONBAR)) redraw_client_frame(c);
        } else {
          c->osxwinfo.view_state=VIEWSTATE_DOCKED;
          if(c->osxwinfo.winref) ChangedOSXWinState(&(c->osxwinfo),c->name,c->clientgroup);
        }
      } else {
        XGetWindowAttributes(dpy, windows[i], &attr);
//      if(attr.map_state==IsViewable) {
        x = attr.x;
        y = attr.y;
//      }
        XMoveWindow(dpy, windows[i], HACKCMD_STARTX, HACKCMD_STARTY);
        XMoveWindow(dpy, windows[i], HACKCMD_HIDEWINDOW, 0);
        XRaiseWindow(dpy, windows[i]);
        if(c->is_docked!=0) {
          c->osxwinfo.view_state=VIEWSTATE_DOCKED;
        } else if(c->is_icon!=NOT_ICONBAR) {
          c->osxwinfo.view_state=VIEWSTATE_ICONBAR;
        } else {
          set_wm_state(windows[i], NormalState);
          c->osxwinfo.view_state=VIEWSTATE_VISIBLE;
        }
        if(c->has_frame) redraw_client_frame(c);
        XMoveWindow(dpy,windows[i],FX,FY);
        XFlush(dpy);
        if(c->osxwinfo.winref) ChangedOSXWinState(&(c->osxwinfo),c->name,c->clientgroup);
//      } else {
//        set_wm_state(windows[i], NormalState);
//        XMoveWindow(dpy, windows[i], x, y);
//        XFlush(dpy);
      }
    }
  } 

  if(windows) XFree(windows);

  MyDecreaseXServerGrab(); // release the X server, if it was before

}

void unmap_group(long group_winref) {
  Window *windows=NULL, w1, w2; 
  int count, i, x, y; 
  Client *c;
  XWindowAttributes attr;
  
//  focus_before_hide=get_focus_client();
  
//  if(can_send_hackcmds!=0) {
//    SendSimpleHackCommand(HACKCMD_HIDE_ALLXWIN,1);
//    return;
//  }
  
    if(client_count==0) return;
    if(client_list==NULL) return;
    
    if(gVerboseConsole)
        printf("%s info: unmap_group with winref %#lx\n",PROGNAME,group_winref);
    if(group_winref==0) return;

  sprintf(lastsub,"TRACE1: unmap_group %#lx\n",group_winref); 
  subtrace();
#ifdef SHOW_TRACE3
  printf("TRACE1: unmap_group %#lx\n",group_winref); 
#endif
 
  MyIncreaseXServerGrab(); // grab the server, so nothing happens while we're messing around
  XQueryTree(dpy, root, &w1, &w2, &windows, &count); 
#ifdef DEBUG
  printf("  children of root: %d\n",count);
#endif

  for(i=count-1;i>=0;i--) 
  { 
#ifdef DEBUG
    printf("  child %d (%#lx)\n",i,windows[i]);
#endif
    c=get_client_from_window(windows[i],FRAME);
    if(c!=NULL && get_group_for_client(c)==group_winref) {
      XLowerWindow(dpy,windows[i]);
      if(can_send_hackcmds==0) {
        if(c->is_docked==0) {
          c->ignore_unmap++; 
          XUnmapWindow(dpy, c->frame); 
          c->osxwinfo.view_state=VIEWSTATE_HIDDEN;
          if(c->osxwinfo.winref) ChangedOSXWinState(&(c->osxwinfo),c->name,c->clientgroup);
          XUnmapWindow(dpy, c->window); 
          set_wm_state(c->window, WithdrawnState); 
        }
      } else {
        XGetWindowAttributes(dpy, windows[i], &attr);
        if(attr.map_state==IsViewable) {
          c->osxwinfo.view_state=VIEWSTATE_HIDDEN;
          if(c->osxwinfo.winref) ChangedOSXWinState(&(c->osxwinfo),c->name,c->clientgroup);
          x = attr.x;
          y = attr.y;
          set_wm_state(windows[i], WithdrawnState);
          XMoveWindow(dpy, windows[i], HACKCMD_STARTX, HACKCMD_STARTY);
          XMoveWindow(dpy, windows[i], HACKCMD_HIDEWINDOW, 1);
          XMoveWindow(dpy, windows[i], x, y);
          XFlush(dpy);
        }
      }
    }
  }

  c = get_top_most_client(WIN_LAYER_NORMAL,0);
  if(c!=NULL) set_focus_client(c);
  
  MyDecreaseXServerGrab(); // release the X server, if it was before
  
  if(windows) XFree(windows);
  
//  c=client_list;
//  for(i=0;i<client_count;i++) {
//    c=c->next;
//  }
}
