/*
 * NSView subclass for Mac OS X rootless X server
 */
/* $XFree86: xc/programs/Xserver/hw/darwin/bundle/XView.m,v 1.5 2001/11/09 00:12:38 torrey Exp $ */

#include <ApplicationServices/ApplicationServices.h>

#import "XView.h"
#include "fakeBoxRec.h"
#include "unistd.h"
#include "quartzCommon.h"

extern void ErrorF(const char *, ...);

static const void *infobytes(void *info)
{
    return info;
}


static unsigned long *shapeBits = NULL, *tempBits = NULL;
static int shapeWidth = 0;
static int shapeHeight = 0;
static int shapeDrawing = 0;
char lastDrawSet[30];
char lastShapeSet[30];
char lastFlushSet[30];
char lastAbortSet[30];

/*
 * These routines are used to make sure we don't get a deadlock between threads when either the window
 * bits or the shapeBits are in use.
 * They will give up after the supplied time (in 1/100ths sec) is provided (usually 1 second),
 * but they do give a warning first (after 9/10ths second, usually).
 */
static int SleepingForShape(int hid, int fivehundths, int warn, char *name)
{
    int i;
    
    i=0;
    while(shapeDrawing) {
        usleep(2000);
        i++;
        if(i>fivehundths) {
            NSLog(@"HACKID %d : give up waiting for shapeBits in %s ****",hid,name);
            if(shapeDrawing) return 1;
        } else if(i>warn)
            NSLog(@"hackid %d : sleep in %s until shapeBits freed (%s)",hid,name,lastShapeSet);
    }
    shapeDrawing=1;
    sprintf(lastShapeSet,"%s",name);
    return 0;
}

static void reallocShapeBits(int id, NSSize minSize)
{
    unsigned long *oldBits;
    int oldWidth, oldHeight;
    
    if(shapeDrawing==0)
        NSLog(@"HACKID %d WARNING: In reallocShapeBits with shapeDrawing = 0 ***",id);

    if (shapeWidth*shapeHeight < minSize.width*minSize.height) {
//        oldBits=shapeBits;
//        oldWidth=shapeWidth;
//        oldHeight=shapeHeight;
        shapeWidth = minSize.width;
        shapeHeight = minSize.height;
        if (shapeBits) return; //free(shapeBits);
        shapeBits = (unsigned long *) malloc(400500);
//        shapeBits = (unsigned long *) malloc(4 * shapeWidth * shapeHeight);
//        NSLog(@"Reallocated shape bits... %#lx %d x %d (%#lx %d x %d)",shapeBits,shapeWidth,shapeHeight,oldBits,oldWidth,oldHeight);
    }
}


@implementation XView

- (id)initWithFrame:(NSRect)aRect isRoot:(BOOL)isRoot
{
//    mShaped = NO;
    maxStoredCount = 100;
    mStoredList = malloc(sizeof(fakeBoxRec)*maxStoredCount);
    
    if(mStoredList==NULL) {
        NSLog(@"FAILED to allocate mStoredList!!");
        return nil;
    }
    
    self = [super initWithFrame:aRect];
    if (!self) return nil;

    mWaitingForTimer = NO;
    mAbortRefreshRects = NO;
    mFlushingRects = NO;
    mShaped = YES;
    mBitsPerSample = 8;
    mSamplesPerPixel = 3;
    mDepth = mBitsPerSample * mSamplesPerPixel;
    mBitsPerPixel = 32;
    mRows = 0;
    mBits = nil;
    mDrawing = YES;
    [self allocBitsForSize:aRect.size isRoot:isRoot];
    mDrawing = NO;
    mDimmed = NO;
    mInhibitShadowChange = NO;
    mMiniature = NO;
    mDrawAlpha = 0xff000000;
    mTranslucentBG = 0xff909090;
    mWaitForDrawRect = NO;
//    mInhibitUpdate = NO;
//    mInhibitRefresh = NO;
// this will be filled in with an ID from OroborOSX, so XDarwin can signal a specific window
    if(isRoot)
        mHackID = -1; // special ID for root window
    else
        mHackID = 0;
    myFrame=aRect;
    mInhibitLevel=0;
    //[self setAutoresizingMask:(NSViewMinYMargin | NSViewMaxXMargin)];
    //[self setAutoresizingMask:(NSViewMinYMargin | NSViewMaxXMargin)];
    
    return self;
}

-(int)SleepingForBitsShape:(int)fivehundths warn:(int)warn who:(char *)name
{
    int i;
    
    i=0;
    while(shapeDrawing || mDrawing) {
        usleep(2000);
        i++;
        if(i>fivehundths) {
            NSLog(@"HACKID %d : give up waiting for shape and bits in %s (%#lx) ****",mHackID,name,mBits);
            if(shapeDrawing || mDrawing) return 1;
        } else if(i>warn) {
            if(shapeDrawing) NSLog(@"hackid %d : sleep in %s until shapeBits freed (%#lx) (%s)",mHackID,name,mBits,lastShapeSet);
            if(mDrawing) NSLog(@"hackid %d : sleep in %s until mBits freed (%#lx) (%s)",mHackID,name,mBits,lastDrawSet);
        }
    }
    shapeDrawing=1;
    mDrawing = YES;
    sprintf(lastShapeSet,"%s",name);
    sprintf(lastDrawSet,"%s",name);
    return 0;
}

-(int)SleepingForBits:(int)fivehundths warn:(int)warn who:(char *)name
{
    int i;
    
    i=0;
    while(mDrawing) {
        usleep(2000);
        i++;
        if(i>fivehundths) {
            NSLog(@"HACKID %d : give up waiting for bits in %s (%#lx) ****",mHackID,name,mBits);
            if(mDrawing) return 1;
        } else if(i>warn) {
            NSLog(@"hackid %d : sleep in %s until mBits freed (%#lx) (%s)",mHackID,name,mBits,lastDrawSet);
        }
    }
    mDrawing = YES;
    sprintf(lastDrawSet,"%s",name);
    return 0;
}

-(int)WaitForFlushingRects:(int)fivehundths warn:(int)warn who:(char *)name
{
    int i;
    
    i=0;
    while(mFlushingRects) {
        usleep(2000);
        i++;
        if(i>fivehundths) {
            NSLog(@"HACKID %d : give up waiting for flush in %s (%d) ****",mHackID,name,mStoredCount);
            if(mFlushingRects) return 1;
        } else if(i>warn) {
            NSLog(@"hackid %d : sleep in %s until flushed (%d) (%s)",mHackID,name,mStoredCount,lastDrawSet);
        }
    }
    sprintf(lastFlushSet,"%s",name);
    return 0;
}

- (void)viewClosing
{
    mChangingRectList = YES;
    // check if the main thread is flushing the rect list
    if(mFlushingRects) {
        sprintf(lastAbortSet,"viewClosing");
        mAbortRefreshRects = YES;
        [self WaitForFlushingRects:500 warn:200 who:"viewClosing"];
        mAbortRefreshRects = NO;
    }
    mStoredCount = 0;
    if(mStoredList) free(mStoredList);
    mStoredList = NULL;
    mChangingRectList = NO;
}

- (void)dealloc
{
    [self SleepingForBits:500 warn:200 who:"dealloc"];
    if (mBits) free(mBits);
    if(mStoredList) free(mStoredList);
    mBits = NULL;
    mStoredList = NULL;
    [super dealloc];
}
/*
- (void)copyAllToShapeBits
{
    fakeBoxRec bounds = {[self frame].origin.x, [self frame].origin.y, [self frame].size.width,
                         [self frame].size.height};
    [self copyToShapeBits:&bounds count:1];
}
- (void)lockFocus
{
    NSLog(@"HackID %d: lockFocus",mHackID);
    [super lockFocus];
}

- (void)unlockFocus
{
    NSLog(@"HackID %d: unlockFocus",mHackID);
    [super unlockFocus];
}
*/

- (void)drawRect:(NSRect)aRect
{
    unsigned long oldAlpha;
    fakeBoxRec bounds = {aRect.origin.x, aRect.origin.y, aRect.size.width,
                         aRect.size.height};
    BOOL	wasMin;
    int		oldInhibit, dummy;
    
    // Never draw here.  - oh yeah? Says who...?!
    //NSLog(@"HACKID %d : Wants to draw... %#lx %d %d",mHackID,mBits,mWaitForDrawRect,mMiniature);
    
//    return;
    if(mStoredList==NULL) {
        NSLog(@"Ignoring drawRect with mStoredList=NULL");
        return;
    }
    
    if(!mMiniature && !mWaitForDrawRect) return;
    
    if(mHackID==-1) {
        NSLog(@"ignoring drawRect for root window");
        return;
    }

    wasMin = mMiniature;
    
        [self SleepingForBitsShape:500 warn:200 who:"drawRect"];
//        reallocShapeBits(mHackID,aRect.size); // don't do this here since it should be already done now
        mMiniature=NO; // temporarily allow copyToScreen to draw to window
        // copy everything from X11 to temp
        // copy everything from temp to screen
//        [self copyToShapeBits:&bounds count:1];
//        if(!mWaitForDrawRect || mDrawAlpha==0xffffffff) {
//            NSLog(@"HACKID %d : lockFocus in drawRect... %#lx",mHackID,mBits);   
            [self lockFocus];
            [[NSColor clearColor] set];
            NSRectFill(aRect);
            [self unlockFocus];
//        }
        // make sure it does complete screen draw of entire window
        oldAlpha = mDrawAlpha;
        mDrawAlpha = 0xffffffff;
        oldInhibit = mInhibitLevel;
        mInhibitLevel = 0;
//        NSLog(@"HACKID %d : lockFocus2 in drawRect... %#lx",mHackID,mBits);   
        [self lockFocus];
//        ErrorF("HACKID %d : drawRect doing a rect with mDrawAlpha=%#lx and mDimmed=%d\n",mHackID,mDrawAlpha,mDimmed);
        //NSLog(@"HACKID %d : doing copy in drawRect... %#lx",mHackID,mBits);   
        dummy = [self copyToScreenViaShape:&bounds count:0];
//        [self copyToScreen:&bounds count:1 fromTemp:YES];
        //NSLog(@"HACKID %d : done copy in drawRect... %#lx",mHackID,mBits);   
        [[NSGraphicsContext currentContext] flushGraphics];
        [self unlockFocus];
        mInhibitLevel = oldInhibit;
        mDrawAlpha = oldAlpha;
        mMiniature=wasMin;
        shapeDrawing = 0;
        mDrawing = NO;
        if(mWaitForDrawRect && mMiniature && mInhibitLevel>=1000) mWaitForDrawRect=NO;
//        [[self window] resetAlphaValue];
//        mWaitForDrawRect=NO;
        //NSLog(@"HACKID %d : done drawRect... %#lx",mHackID,mBits);   
}

- (void)setWaitForDrawRect:(BOOL)val
{
    mWaitForDrawRect = val;
}

- (BOOL)getWaitForDrawRect
{
    return mWaitForDrawRect;
}

- (BOOL)isFlipped
{
    return NO;
}

- (BOOL)isOpaque
{
//    return NO;
    return YES;
//    return !mShaped;
}

- (BOOL)acceptsFirstResponder
{
    return !mMiniature;
}

- (BOOL)acceptsFirstMouse:(NSEvent *)theEvent
{
    return !mMiniature;
}

- (BOOL)shouldDelayWindowOrderingForEvent:(NSEvent *)theEvent
{
    return !mMiniature;
}


- (void)mouseEntered:(NSEvent *)anEvent
{
    xwinHackID[hackIDPos] = mHackID;
    if(verboseConsole)
        NSLog(@"HACKID %d : <<<<<<<< Sending entered notification\n",mHackID);
    QuartzMessageMainThread(kQuartzWinEntered+hackIDPos);
    hackIDPos++;
    if(hackIDPos>=kDockPosMax) hackIDPos=0;
}

- (void)mouseExited:(NSEvent *)anEvent
{
    xwinHackID[hackIDPos] = mHackID;
    if(verboseConsole)
        NSLog(@"HACKID %d : <<<<<<<< Sending exited notification\n",mHackID);
    QuartzMessageMainThread(kQuartzWinExited+hackIDPos);
    hackIDPos++;
    if(hackIDPos>=kDockPosMax) hackIDPos=0;
}

- (void)mouseDown:(NSEvent *)anEvent
{
//    if(mMiniature) return;
    // Only X is allowed to restack windows.
    if(verboseConsole)
        NSLog(@"HACKID %d : Got mouse down... (%d)",mHackID,can_come_in_front);
    [NSApp preventWindowOrdering];
    if (! [NSApp isActive] && can_come_in_front==1)	// used when OroborOSX is in menu select
        [NSApp activateIgnoringOtherApps:YES];
    [[self window] setAcceptsMouseMovedEvents:YES];
    [[self nextResponder] mouseDown:anEvent];
}

- (void)mouseUp:(NSEvent *)anEvent
{
    if(verboseConsole)
        NSLog(@"HACKID %d : Got mouse up... (%d)",mHackID,can_come_in_front);
//    if(mMiniature) return;
    // Bring app to front if necessary
    // Don't bring app to front in mouseDown; mousedown-mouseup is too
    // long and X gets what looks like a mouse drag.
    if (! [NSApp isActive] ) {
        if(can_come_in_front==2)	// not currently used
            [NSApp activateIgnoringOtherApps:YES];
        else if(can_come_in_front==3) {	// original XDarwin behaviour
            [NSApp activateIgnoringOtherApps:YES];
            [NSApp arrangeInFront:nil];
        }
    }

    [[self nextResponder] mouseDown:anEvent];
//    [[self nextResponder] mouseUp:anEvent];
}

//- (NSRect)frame
//{
//    return myFrame;
//}

//- (void)setFrame:(NSRect)aRect
//{
//    fakeBoxRec bounds = {0, 0, aRect.size.width, aRect.size.height};

//    if(mHackID>0) {
//        NSLog(@"HACKID %d : removing tracking rect in setFrame... %d",mHackID,mRectTag);   
//        [self removeTrackingRect:mRectTag];
//    }
//    myFrame=aRect;
//    [super setFrame:aRect];
//    if(mHackID>0) {
//        mRectTag = [self addTrackingRect:[self frame] owner:self userData:nil assumeInside:NO];
//        NSLog(@"HACKID %d : changed tracking rect in setFrame... %d",mHackID,mRectTag);
//    }
//}

// setFrame goes through here too.
//- (void)setFrameSize:(NSSize)newSize
//{
//    fakeBoxRec bounds = {0, 0, newSize.width, newSize.height};
    
/*    if(mHackID>0) {
        NSLog(@"HACKID %d : removing tracking rect in setFrameSize... %d",mHackID,mRectTag);   
        [self removeTrackingRect:mRectTag];
    }*/
//    [self allocBitsForSize:newSize];
//        [self lockFocus];
//        [[NSColor clearColor] set];
//        NSRectFill([self frame]);
//        [self unlockFocus];
//    [super setFrameSize:newSize];
    //myFrame.size=newSize;
/*    if(mHackID>0) {
        mRectTag = [self addTrackingRect:[self frame] owner:self userData:nil assumeInside:NO];
        NSLog(@"HACKID %d : changed tracking rect in setFrameSize... %d",mHackID,mRectTag);
    }*/
    
        // copy everything from X11 to temp
        // copy everything from temp to screen
/*	if(shapeDrawing) SleepingForShape(100,90,"setFrameSize");
//        [self lockFocus];
        [self copyToShapeBits:&bounds count:1];
        [self copyToScreen:&bounds count:1 fromTemp:YES];
        [[NSGraphicsContext currentContext] flushGraphics];
//        [self unlockFocus];
        shapeDrawing = 0;
*/
//}

// Reallocate bits.
- (void)allocBitsForSize:(NSSize)newSize isRoot:(BOOL)isRoot
{
    unsigned long *pix;
    int i;
    
    if(!mDrawing && !isRoot)
        NSLog(@"HACKID %d WARNING: In allocBitsForSize with mDrawing = NO ***",mHackID);

    mBytesPerRow = newSize.width * mBitsPerPixel / 8;
    mRows = newSize.height;

        pix=mBits;
        if(mBits!=nil) {
//            free(mBits);
            mBits = realloc(mBits,mBytesPerRow * newSize.height);
//            NSLog(@"HACKID %d : reallocated mBits: %#lx (%#lx)",mHackID,mBits,pix);
            ;
        } else if(!isRoot) {
//            mBits = malloc(4000000);
            mBits = malloc(mBytesPerRow * newSize.height);
//            NSLog(@"HACKID %d : newly allocated mBits: %#lx",mHackID,mBits);
            pix = mBits;
            for(i=0;i<newSize.width*newSize.height;i++)
                *pix++ = 0;
            SleepingForShape(mHackID,500,200,"allocBitsForSize");
            reallocShapeBits(mHackID,newSize);
            shapeDrawing=0;
        } else {
            mBits = malloc(mBytesPerRow * newSize.height);
        }
}

- (void)checkAndAllocBitsForSize:(NSSize)newSize isRoot:(BOOL)isRoot
{
    if(!isRoot) [self SleepingForBits:500 warn:200 who:"checkAndAllocBitsForSize"];
    [self allocBitsForSize:newSize isRoot:isRoot];
    mDrawing = NO;
}

- (char *)bits
{
    return mBits;
}

- (void)getBits:(char **)bits
       rowBytes:(int *)rowBytes
          depth:(int *)depth
   bitsPerPixel:(int *)bpp
{
    *bits = mBits;
    *rowBytes = mBytesPerRow;
    *depth = mDepth;
    *bpp = mBitsPerPixel;
}

- (void)flushStored:(NSTimer *)aTimer
{

    mWaitingForTimer = NO;
    
    if(mStoredList==NULL) {
        if(aTimer) [aTimer invalidate];
        return;
    }
    
    mFlushingRects = YES;
    if(mStoredCount==0) {
        if(aTimer) [aTimer invalidate];
        mFlushingRects = NO;
        return;
    }

    if(mChangingRectList && aTimer) { // if aTimer is nil then force the flush
        [aTimer invalidate];
        if(verboseConsole)
            NSLog(@"------------------- starting new timer for HackID %d...",mHackID);
        [self startTimer];
        mFlushingRects = NO;
        return;
    }
    
    if(verboseConsole)
        NSLog(@"================= flushing rects list (%d) for HackID %d",mStoredCount,mHackID);
    if(aTimer) [aTimer invalidate];
    if(mStoredCount>0)
        [self refreshRects:mStoredList count:mStoredCount];
    mStoredCount=0;
    mFlushingRects = NO;
}

- (void)storeRects:(fakeBoxRec *)rectList count:(int)count
{
    fakeBoxRec *r, *r2;
    fakeBoxRec *cur;
    fakeBoxRec *end, *end2;
    NSRect aRect;
    int smallcnt=0;
    fakeBoxRec *smallrcts;

    if(mStoredList==NULL) {
        NSLog(@"Ignoring storeRects with mStoredList=NULL");
        return;
    }
    
    if(mHackID==-1) {
        NSLog(@"ignoring storeRects (%d) for root window",count);
        return;
    }
    
    if(count<=0) {
        NSLog(@"HackID %d : *** Got into storeRects with zero rects (%d)!!!",mHackID,count);
        return;
    }
    
    if(mInhibitLevel>10 && count!=0 && mHackID>0 && mDrawAlpha==0xff000000) {
        //NSLog(@"HackID %d : ignoring %d rects for refreshRects",mHackID,count);
        return;
    } else if(mMiniature) {
        NSLog(@"HackID %d : ignoring %d rects for refresh of miniwin",mHackID,count);
        return;
    } else if(mAbortRefreshRects) {
        NSLog(@"+++++++++ Aborting storeRects (%d + %d, \"%s\")",count,mStoredCount,lastAbortSet);
        return;
    }
/*
    [self refreshRects:rectList count:count];
    mStoredCount=0;
            
    return;
*/
    mChangingRectList = YES;
    // check if the main thread is flushing the rect list
    if(mFlushingRects) {
//        NSLog(@"Waiting for flush of %d rects in storeRects (%d)...",mStoredCount,count);
        [self WaitForFlushingRects:500 warn:200 who:"storeRects"];
        // might as well finish off the rest of the updates  (...not!)
//        [self refreshRects:rectList count:count];
//        mChangingRectList = NO;
//        return;
    }
    mAbortRefreshRects = NO;
    
    if(count+mStoredCount>=maxStoredCount) {
        if(verboseConsole)
            NSLog(@"HackID %d : *** Too many rects in storeRects (%d + %d)",mHackID,count,mStoredCount);
        if(mStoredCount>0) [self refreshRects:mStoredList count:mStoredCount];
        mStoredCount=0;
        while (count>=maxStoredCount) {
            [self refreshRects:rectList count:(maxStoredCount-1)];
            count -= (maxStoredCount - 1);
            rectList += (maxStoredCount - 1);
        }
//        mChangingRectList = NO;
//        return;
    }
    
    if(verboseConsole)
        NSLog(@"HackID %d : Storing %d more rects in storeRects (%d)",mHackID,count,mStoredCount);
    cur = mStoredList + mStoredCount;
    for (r = rectList, end = rectList + count; r < end; r++) {
        for(r2 = mStoredList, end2 = mStoredList+mStoredCount; r2 < end2; r2++) {
            if((r->x1 >= r2->x2) || (r->y1 >= r2->y2) || (r2->x1 >= r->x2) || (r2->y1 >= r->y2)) {
                ;
            } else if(r->x1 <= r2->x1) {
                if(r->y1 <= r2->y1) {
                    if(r->y2 >= r2->y2) {
                        if(r->x2 >= r2->x2) {
                            *r2 = *r;	// new rect contains old, so replace old with new
                            r->x2 = r->x1;	// make sure no new rect is added
                            break;
                        } else {
                           r2->x1 = r->x2; // chop off left of old rect
                        }
                    } else if(r->x2 >= r2->x2) {
                        r2->y1 = r->y2;	// chop off top of old rect
                    }
                } else if(r->x2 >= r2->x2) {
                    if(r->y2 >= r2->y2) {
                        r2->y2 = r->y1;	// chop off bottom of old rect
                    }
                } else if(r->y2 <= r2->y2) {
                    r->x2 = r2->x1;	// chop off right of new rect
                }
            } else if(r->y1 >= r2->y1) {
                if(r->y2 <= r2->y2) {
                    if(r->x2 <= r2->x2) {
                        r->x2 = r->x1;	// new rect contained in old, so ignore it
                        break;
                    } else {
                       r->x1 = r2->x2;	// chop off left of new rect
                    }
                } else if(r->x2 <= r2->x2) {
                    r->y1 = r2->y2;	// chop off top of new rect
                }
            } else if(r->x2 <= r2->x2) {
                if(r->y2 <= r2->y2) {
                    r->y2 = r2->y1;	// chop off bottom of new rect
                }
            } else if(r->y1 >= r2->y2) {
                r2->x2 = r->x1;	// chop off right of old rect
            }
        }
        if((r->x2 > r->x1) && (r->y2 > r->y1)) {
//            if(rh<30 || rw<30 || rh*rw<10000) {
//                smallrcts--;
//                smallcnt++;
//                *smallrcts = *r;
//            } else {
                *cur = *r;
                cur++;
                mStoredCount++;
//            }
        }
    }
    
    smallcnt = 0;
    smallrcts = mStoredList + maxStoredCount;
    end = mStoredList + mStoredCount;
    for (r = mStoredList; r < end; r++) {
        int done = 0;
        while(done==0) {
            int rh = r->y2 - r->y1;
            int rw = r->x2 - r->x1;
            // is this a small rect?
            if(rh<30 || rw<30 || rh*rw<10000) {
                // put small rects at the end of the rect list
                // (should be safe since above we allowed total rects to only reach one before end)
                smallrcts--;
                smallcnt++;
                *smallrcts = *r;
                end--;
                // check we have not got to the end
                if(r==end)
                    done=1;
                else
                    *r = *end;
                mStoredCount--;
            } else
                done = 1;
        }
    }
    
    if(smallcnt>0) [self refreshRects:smallrcts count:smallcnt];	// draw small rects immediately
    mChangingRectList = NO;
}

- (int)getStoredCount
{
    return mStoredCount;
}

- (void)startTimer
{
    NSTimer *aTimer;
    
    if(mStoredList==NULL) {
        NSLog(@"Ignoring startTimer with mStoredList=NULL");
        return;
    }
    
    mWaitingForTimer = YES;
    
//    NSLog(@"------------------- starting a timer for HackID %d...",mHackID);
    aTimer = [NSTimer scheduledTimerWithTimeInterval:0.05 target:self
            selector:@selector(flushStored:) userInfo:nil repeats:NO];
//        [[NSRunLoop currentRunLoop] addTimer: aTimer forMode: NSEventTrackingRunLoopMode];
}

- (void)refreshRects:(fakeBoxRec *)rectList count:(int)count
{
    fakeBoxRec *r;
    fakeBoxRec *end;
    NSRect aRect;
    int dummy;

    if(mStoredList==NULL) {
        NSLog(@"Ignoring refreshRects with mStoredList=NULL");
        return;
    }
    
    if(mHackID==-1) {
        NSLog(@"ignoring refreshRects (%d) for root window",count);
        return;
    }
    if(count==0) {
        mChangingRectList = YES;
        // check if the main thread is flushing the rect list
        if(mFlushingRects) {
            sprintf(lastAbortSet,"refreshRects");
            mAbortRefreshRects = YES;
            [self WaitForFlushingRects:500 warn:200 who:"refreshRects"];
            mAbortRefreshRects = NO;
        }
        mStoredCount=0;
        mChangingRectList = NO;
    }
//    if(mMiniature) return;
//    if(count==0) 
//        NSLog(@"HackID %d : In refreshRects with no rects!",mHackID);

//    if(mInhibitRefresh && !mMiniature && count!=0)
//        NSLog(@"HackID %d : inhibited %d rects refresh",mHackID,count);
//    if(mInhibitRefresh && count!=0 && mHackID!=0 && mDrawAlpha==0xff000000) {
/*    if(mInhibitLevel>10 && count!=0 && mHackID>0 && mDrawAlpha==0xff000000) {
        //NSLog(@"HackID %d : ignoring %d rects for refreshRects",mHackID,count);
        return;
    } else if(mMiniature) {
        NSLog(@"HackID %d : ignoring %d rects for refresh of miniwin",mHackID,count);
        return;
    }*/
//    NSLog(@"HackID %d : %d rects in refreshRects",mHackID,count);
    
/*    if (!mShaped) {
        [self lockFocus];
        [self copyToScreen:rectList count:count fromTemp:NO];
        [[NSGraphicsContext currentContext] flushGraphics];
        [self unlockFocus];
    } else {*/
        if(mAbortRefreshRects && count>0) {
            NSLog(@"Getting out of refreshRects before anything... (%d, \"%s\")",count,lastAbortSet);
            return;
        }
        [self SleepingForBitsShape:500 warn:200 who:"refreshRects"];
        dummy = count;
        
//        if(verboseConsole)
//            NSLog(@"HackID %d : in refreshRects (%d)",mHackID,count);
        if(!mAbortRefreshRects || count==0) {
        // only need to clear if the drawing might not be completely opaque
            if((mDrawAlpha!=0xff000000 || (mTranslucentBG & 0xff000000)!=0xff000000) && mInhibitLevel<10) {
//            NSLog(@"HACKID %d : lockFocus in refreshRects... %d",mHackID,count);
                [self lockFocus];
                for (r = rectList, end = rectList + count; r < end; r++) {
                    NSRect nsr = {{r->x1, r->y1}, {r->x2-r->x1, r->y2-r->y1}};
                    [[NSColor clearColor] set];
                    nsr.origin.y = [self frame].size.height - nsr.origin.y - nsr.size.height;
//                mWaitForDrawRect=YES;
//                [[self window] setAutodisplay:YES];
//                [self setNeedsDisplayInRect:nsr];
//                while(mWaitForDrawRect) {
//                    NSLog(@"Waiting for drawRect in refreshRects");
//                    usleep(100);
//                }
//                [[self window] setAutodisplay:NO];
                    NSRectFill(nsr);
                }
//            return;
                [self unlockFocus];
            }
        
//        NSLog(@"HACKID %d : lockFocus2 in refreshRects... %d",mHackID,count);
            if(mInhibitLevel<1000) [self lockFocus];
//        for (r = rectList, end = rectList + count; r < end; r++) {
//            [self copyToShapeBits:r count:1];
//            [self copyToScreen:r count:1 fromTemp:YES];
//        }
//        ErrorF("HACKID %d : refreshRects doing %d rects with mDrawAlpha=%#lx and mDimmed=%d\n",mHackID,count,mDrawAlpha,mDimmed);
            dummy = [self copyToScreenViaShape:rectList count:count];
            if(mInhibitLevel<1000) {
                [[NSGraphicsContext currentContext] flushGraphics];
                [self unlockFocus];
            }
        }
        if( dummy > 0 ) NSLog(@"Aborted %d refreshRects (%d left \"%s\")",count,dummy,lastFlushSet);
        
        shapeDrawing = 0;
        mDrawing = NO;
//    }
//        NSLog(@"HackID %d : finished refreshRects",mHackID);
}

// eraseRects are OUTSIDE the new shape
- (void)reshapeRects:(fakeBoxRec *)eraseRects count:(int)count
{
    unsigned long oldAlpha;
    int oldInhibit, dummy;
    
    fakeBoxRec bounds = {0, 0, [self frame].size.width,
                         [self frame].size.height};
    fakeBoxRec *r;
    fakeBoxRec *end;
    NSRect aRect;

    if(mStoredList==NULL) {
        NSLog(@"Ignoring reshapeRects with mStoredList=NULL");
        return;
    }
    
    if(mHackID==-1) {
        NSLog(@"ignoring reshapeRects (%d) for root window",count);
        return;
    }

//    if(mMiniature) return;
// Hmmm... does actually get into here with zero rects for some menus (eg nedit)
// Makes it not update the entire menu window contents if it just returns...
//    if(count==0) 
//        if(mInhibitRefresh && !mMiniature && count!=0 && mHackID>0) {
//        if(mInhibitLevel>10 && !mMiniature && count!=0 && mHackID>0) {
            //NSLog(@"HackID %d : not ignoring reshape of %d rects",mHackID,count);
//            return;
//        } else
            //NSLog(@"HackID %d : reshaping with %d rects",mHackID,count);
    
/*    if (!mShaped) {
        [self refreshRects:&bounds count:1];
    } else {*/
        // View is shaped, or used to be shaped.
        // Shaped windows never become unshaped.
        // (Mac OS X 10.0.4 does not allow transparent windows
        // to become opaque.)

        mShaped = YES;

        mChangingRectList = YES;
        // check if the main thread is flushing the rect list
        if(mFlushingRects) {
            sprintf(lastAbortSet,"reshapeRects");
            mAbortRefreshRects = YES;
            [self WaitForFlushingRects:500 warn:200 who:"reshapeRects"];
            mAbortRefreshRects = NO;
        } else if(mStoredCount>0) {
            if(verboseConsole)
                NSLog(@"HackID %d : flushing rects in reshapeRects (%d)",mHackID,count);
            [self flushStored:nil];	// send nil to force flush, despite mChangingRectList being set
        }
        mChangingRectList = NO;
//        SleepingForShape(mHackID,500,200,"reshapeRects1");
//        reallocShapeBits(mHackID,[self frame].size);
//        shapeDrawing = 0;
        // copy everything from X11 to temp
        // erase eraseRects from temp
        // copy everything from temp to screen
        [self SleepingForBits:500 warn:200 who:"reshapeRects2"];
        [self eraseFromShapeBits:eraseRects count:count];  // now uses the actual window bits
        
//        if(!mInhibitUpdate && !mMiniature) {
        if(mInhibitLevel<=10 && !mMiniature) {
//        if(!mMiniature) {
            if(mInhibitLevel<10) {	// only update actual screen on inhibitlevel<10
                // Magic. 10.0.4 and 10.1 both require the alpha channel to be
                // cleared explicitly. 10.0.4 additionally requires the view to
                // be unlocked between this and the drawing code below.
//                NSLog(@"HACKID %d : lockFocus in reshapeRects... %d",mHackID,count);
                [self lockFocus];
                //[[NSColor blackColor] set];
                //NSRectFill([self frame]);
                for (r = eraseRects, end = eraseRects + count; r < end; r++) {
                    NSRect nsr = {{r->x1, r->y1}, {r->x2-r->x1, r->y2-r->y1}};
                    [[NSColor clearColor] set];
                    nsr.origin.y = [self frame].size.height - nsr.origin.y - nsr.size.height;
                    NSRectFill(nsr);
                }
                [self unlockFocus];
            }

            mDrawing = NO;
//        return;
//        [self setNeedsDisplay:YES];
//        if(!mInhibitUpdate || mMiniature) {
            //NSLog(@"Updating whole window for shape");
//            [self copyToShapeBits:&bounds count:1];
            [self SleepingForBitsShape:500 warn:200 who:"reshapeRects3"];
            oldAlpha = mDrawAlpha;
            mDrawAlpha = 0xffffffff;
//            NSLog(@"HACKID %d : lockFocus2 in reshapeRects... %d",mHackID,count);
            if(mInhibitLevel<1000) [self lockFocus];
//            [self copyToScreen:&bounds count:1 fromTemp:YES];
//            ErrorF("HACKID %d : reshapeRects doing all with mDrawAlpha=%#lx and mDimmed=%d\n",mHackID,mDrawAlpha,mDimmed);
            dummy = [self copyToScreenViaShape:&bounds count:0];
            if(mInhibitLevel<1000) {
                [[NSGraphicsContext currentContext] flushGraphics];
                [self unlockFocus];
            }
            mDrawAlpha = oldAlpha;
            //}
        } else {
            //NSLog(@"Not updating whole window for shape");
            shapeDrawing = 0;
            mDrawing = NO;
            return;
        }
        shapeDrawing = 0;
        mDrawing = NO;
//    }
        //NSLog(@"HackID %d : finished reshaping",mHackID);
}

/*
 * Doesn't use shapeBits any more, but stores transluceny in the free high byte of the window bits
 * warning: may not be a good thing to do!!
 */
- (void)eraseFromShapeBits:(fakeBoxRec *)rectList count:(int)count
{
    unsigned long *bdst = NULL; //  *dst = NULL, don't assign yet
    int dstWidth = 0; // don't assign yet
    fakeBoxRec *r;
    fakeBoxRec *end;

    if(mStoredList==NULL) {
        NSLog(@"Ignoring eraseFromShapeBits with mStoredList=NULL");
        return;
    }
    
//    if(mMiniature) return;

    assert(mBitsPerPixel == 32);

//    reallocShapeBits(mHackID,[self frame].size);
//    dst = shapeBits;
    bdst = (unsigned long *)mBits;
//    dstWidth = shapeWidth;
    dstWidth = mBytesPerRow / 4;

    for (r = rectList, end = rectList + count; r < end; r++) {
//        unsigned long *dstline = dst + dstWidth*r->y1 + r->x1;
        unsigned long *bdstline = bdst + dstWidth*r->y1 + r->x1;
        int h = r->y2 - r->y1;

        while (h--) {
//            unsigned long *dstp  = dstline;
            unsigned long *bdstp = bdstline;
            int w = r->x2 - r->x1;

            while (w--) {
  //              *dstp++  = 0x00000000;
                *bdstp++ = 0xff000000;
            }
//            dstline  += dstWidth;
            bdstline += dstWidth;
        }
    }
}
/*
// assumes X11 bits and shape bits are 32-bit
- (void)copyToShapeBits:(fakeBoxRec *)rectList count:(int)count
{
    unsigned long *src = (unsigned long *) mBits;
    unsigned long *dst = NULL; // don't assign yet
    unsigned long newval;
    int srcWidth = mBytesPerRow / 4;
    int dstWidth = 0; // don't assign yet
    fakeBoxRec *r;
    fakeBoxRec *end;

        NSLog(@"HACKID %d : GOT copyToShapeBits ***",mHackID);
        return;
//    if(mMiniature) return;
    if(count<=0) return;
    
    assert(mBitsPerPixel == 32);

    if(shapeDrawing==0)
        NSLog(@"HACKID %d WARNING: In copyToShapeBits with shapeDrawing = 0 ***",mHackID);
    if(!mDrawing)
        NSLog(@"HACKID %d WARNING: In copyToShapeBits with mDrawing = NO ***",mHackID);
//    reallocShapeBits(mHackID,[self frame].size);
    dst = shapeBits;
//    dstWidth = shapeWidth;
    if(count==1) {
        dstWidth = rectList->x2 - rectList->x1;
        if(dstWidth>srcWidth)
            dstWidth=srcWidth;
    } else
        dstWidth = srcWidth;
    
    for (r = rectList, end = rectList + count; r < end; r++) {
        unsigned long *srcline = src + srcWidth*r->y1 + r->x1;
        unsigned long *dstline = dst + dstWidth*r->y1 + r->x1;
        int h = r->y2 - r->y1;
        if(h>mRows)
            h=mRows;
        if(count==1) dstline = dst;

        while (h--) {
            unsigned long *srcp = srcline;
            unsigned long *dstp = dstline;
            int w = r->x2 - r->x1;
            if(w>dstWidth)
                w=dstWidth;
                
            if(mDimmed) {
                while(w--) {
                    newval = *srcp++;
                    if(newval==0xff000000)
                        *dstp++ = 0;
                    else {
                    newval = newval & 0x00ffffff;
                    if(newval==(mTranslucentBG & 0x00ffffff)) {
                        *dstp++ = mTranslucentBG;
                    } else
                        *dstp++ = (((newval & 0x00fefefe) + (mTranslucentBG & 0x00fefefe)) >> 1) | mDrawAlpha;
                    }
                }
//                    newval = *srcp++ & 0x00fefefe;
//                    newval = (newval << 2) + (newval & 0x00f8f8f8);
//                    *dstp++ = ((newval + 0x01818180) >> 3) | 0xff000000;
            } else {
                while(w--) {
                    newval = *srcp++;
                    if(newval==0xff000000)
                        *dstp++ = 0;
                    else {
                    newval = newval & 0x00ffffff;
                    if(newval==(mTranslucentBG & 0x00ffffff)) {
                        *dstp++ = mTranslucentBG;
                    } else
                        *dstp++ = newval | mDrawAlpha;
                    }
                }
            }
//            while (w--) {
//                *dstp++ = (((*srcp++ & 0x00fefefe)+0x00808080) >> 1) | 0xff000000;
//                newval = (*srcp & 0x00ffffff) | 0xc0000000;
//                if(*dstp != newval)
//                    *dstp++ = (*srcp++ & 0x00ffffff) | 0xc0000000;
//                else {
//                    *dstp++ = 0x00000000;
//                    *srcp++;
//                }
//            }
            srcline += srcWidth;
            dstline += dstWidth;
        }
    }
}


// Copy boxes to the screen from the per-window pixmaps where X draws.
// rectList is in local, X-flipped coordinates.
- (void)copyToScreen:(fakeBoxRec *)rectList count:(int)count
            fromTemp:(BOOL)copyFromTemp
{
    unsigned char *offsetbits;
    fakeBoxRec *r;
    fakeBoxRec *end;
    NSRect bounds;
    char *srcBits;
    int bytesPerRow;
    int bitsPerPixel;
    int bitsPerSample;
    CGImageAlphaInfo alpha;

    CGContextRef destCtx;
    CGColorSpaceRef colorSpace;
    const CGDataProviderDirectAccessCallbacks cb = { infobytes, NULL, NULL, NULL };

    //if( mMiniature ) return;
//    if( mMiniature )
//        NSLog(@"Drawing for miniaturised window (%d)...",mHackID);
        NSLog(@"HACKID %d : GOT copyToScreen ***",mHackID);
        return;
        
    destCtx = (CGContextRef)[[NSGraphicsContext currentContext]
                                                graphicsPort];
                                                
//    destCtx = [NSGraphicsContext currentContext];
    colorSpace = CGColorSpaceCreateDeviceRGB();
    // fixme colorspace leaks?
        
    if (copyFromTemp) {
        // shapeBits assumed to be 32-bit
        srcBits = (char *)shapeBits;
//        bytesPerRow = 4 * shapeWidth;
        if(count==1) {
            bytesPerRow = (rectList->x2 - rectList->x1) * 4;
            if(bytesPerRow>mBytesPerRow)
                bytesPerRow = mBytesPerRow;
        } else
            bytesPerRow = mBytesPerRow;
            
        if(shapeDrawing==0)
            NSLog(@"HACKID %d WARNING: In copyToScreen with shapeDrawing = 0 ***",mHackID);
        bitsPerPixel = 32;
        bitsPerSample = 8;
        alpha = kCGImageAlphaPremultipliedFirst; // premultiplied ARGB (not meant to use -see below)
//        alpha = kCGImageAlphaFirst; // non-premultiplied (should use this but crashes OSX10.0???)
    } else {
        srcBits = mBits;
        bytesPerRow = mBytesPerRow;
        bitsPerPixel = mBitsPerPixel;
        bitsPerSample = mBitsPerSample;
        alpha = kCGImageAlphaNoneSkipFirst; // xRGB
    }

//    bounds = [self frame];
    bounds.origin.x = bounds.origin.y = 0;
    bounds.size.width = mBytesPerRow/4;
    bounds.size.height = mRows;

    for (r = rectList, end = rectList + count; r < end; r++) {
        NSRect nsr = {{r->x1, r->y1}, {r->x2-r->x1, r->y2-r->y1}};
        CGRect destRect;
        CGDataProviderRef dataProviderRef;
        CGImageRef imageRef;

        // Clip to window
        // (bounds origin is (0,0) so it can be used in either flip)
        // fixme is this necessary with pixmap-per-window?
        nsr = NSIntersectionRect(nsr, bounds);

        // Disallow empty rects
        if (nsr.size.width <= 0  ||  nsr.size.height <= 0) continue;

        if(count==1)
            offsetbits = srcBits;
        else
            offsetbits = srcBits + (int)(nsr.origin.y * bytesPerRow +
                                        nsr.origin.x * bitsPerPixel/8);

        // Flip r to Cocoa-flipped
        nsr.origin.y = [self frame].size.height - nsr.origin.y - nsr.size.height;
        destRect = CGRectMake(nsr.origin.x, nsr.origin.y,
                              nsr.size.width, nsr.size.height);

        dataProviderRef = CGDataProviderCreateDirectAccess(offsetbits,
                                destRect.size.height * bytesPerRow, &cb);
//        dataProviderRef = CGDataProviderCreateWithData(NULL,offsetbits,
//                                destRect.size.height * bytesPerRow,NULL);

        imageRef = CGImageCreate(destRect.size.width, destRect.size.height,
                                 bitsPerSample, bitsPerPixel, bytesPerRow,
                                 colorSpace, alpha, dataProviderRef, NULL,
//                                 0, kCGRenderingIntentDefault);
                                 1, kCGRenderingIntentDefault);

        CGContextDrawImage(destCtx, destRect, imageRef);
        CGImageRelease(imageRef);
        CGDataProviderRelease(dataProviderRef);
    }
}
*/

// Combined update routine, in an attempt to make it slightly faster...
// assumes X11 bits and shape bits are 32-bit
- (int)copyToScreenViaShape:(fakeBoxRec *)rectList count:(int)count
{
    unsigned long *src = (unsigned long *) mBits;
    unsigned long *dst = NULL; // don't assign yet
    unsigned long newval;
    int srcWidth = mBytesPerRow / 4;
    int dstWidth = 0, firstDraw=1; // don't assign yet
    fakeBoxRec *r;
    fakeBoxRec *end;
    unsigned char *offsetbits;
    NSRect bounds;
    char *srcBits;
    int bytesPerRow, bitsPerPixel, bitsPerSample, leftRects=0;
    CGImageAlphaInfo alpha;

    CGContextRef destCtx;
    CGColorSpaceRef colorSpace;
    const CGDataProviderDirectAccessCallbacks cb = { infobytes, NULL, NULL, NULL };

    if(mStoredList==NULL) {
        NSLog(@"Ignoring copyToScreenViaShape with mStoredList=NULL");
        return 0;
    }
    
    leftRects=count;
    
    //if( mMiniature ) return;
    if(count==0) {
        count=1; 
        firstDraw=0;
    } else if(count<0)
        return 0;
    else
        firstDraw=1;

//    if( mMiniature )
//        NSLog(@"Drawing for miniaturised window (%d)...",mHackID);
//    if(count==1)
        
    destCtx = (CGContextRef)[[NSGraphicsContext currentContext] graphicsPort];
                                                
//    destCtx = [NSGraphicsContext currentContext];
    colorSpace = CGColorSpaceCreateDeviceRGB();
    // fixme colorspace leaks?
        
    // shapeBits assumed to be 32-bit
    srcBits = (char *)shapeBits;
//    bytesPerRow = mBytesPerRow;
    bitsPerPixel = 32;
    bitsPerSample = 8;
//    alpha = kCGImageAlphaPremultipliedFirst; // premultiplied ARGB (not meant to use -see below)
    alpha = kCGImageAlphaFirst; // non-premultiplied (should use this but crashes OSX10.0???)

    bounds = [self frame];
    bounds.origin.x = bounds.origin.y = 0;
//    bounds.size.width = mBytesPerRow/4;
//    bounds.size.height = mRows;

    if(mMiniature) {
        NSLog(@"HACKID %d WARNING: In copyToScreenViaShape for miniwin! ***",mHackID);
        return 0;
    }
    
    assert(mBitsPerPixel == 32);

    //reallocShapeBits(mHackID,[self frame].size);
    dst = shapeBits;
    
    if(shapeDrawing==0)
        NSLog(@"HACKID %d WARNING: In copyToScreenViaShape with shapeDrawing = 0 ***",mHackID);
    if(!mDrawing)
        NSLog(@"HACKID %d WARNING: In copyToScreenViaShape with mDrawing = NO ***",mHackID);

//    if(mDrawAlpha!=0xff000000)
//        NSLog(@"In copyToScreenViaShape with mDrawAlpha=%#lx (%d %d)",mDrawAlpha,mInhibitLevel,mDimmed);
        
    for (r = rectList, end = rectList + count; r < end; r++) {
    
      int h_left = r->y2 - r->y1;
      int rtop = r->y1;
    
      if( mAbortRefreshRects ) {
        NSLog(@"Aborting copyToScreenViaShape with %d left out of %d (%s)",leftRects,count,lastAbortSet);
        return leftRects;
      }
      
      leftRects--;
      
      if(rtop+h_left>mRows)
        h_left=mRows-rtop;
            
      dstWidth = r->x2 - r->x1;
      bytesPerRow = dstWidth * 4;
      if(bytesPerRow>mBytesPerRow) {
        bytesPerRow = mBytesPerRow;
        dstWidth=bytesPerRow/4;
      }
                
      while( h_left>0 ) {
        
//        NSRect nsr = {{r->x1, r->y1}, {r->x2-r->x1, r->y2-r->y1}};
        NSRect nsr = {{r->x1, rtop}, {dstWidth, h_left}};
        CGRect destRect;
        CGDataProviderRef dataProviderRef;
        CGImageRef imageRef;
        unsigned long *srcline = src + srcWidth*rtop + r->x1;
        unsigned long *dstline = dst;
        int h=h_left;

        // This makes sure that not too much is copied between memory locations in one go.
        // Specifically, it makes sure only at most 20000 pixels of shapeBits is used at once.
        // Hopefully, that will help keep it in cache, so it does not need to go to main
        // memory before the drawing routine uses it...
        if( h_left*dstWidth>100000 )
            h = 100000/dstWidth;
        else if( h_left*dstWidth>50000 )
            h = h_left >> 1;

        nsr.origin.y = rtop;
        nsr.size.height = h;
        rtop+=h;
        h_left-=h;
        // Clip to window
        // (bounds origin is (0,0) so it can be used in either flip)
        // fixme is this necessary with pixmap-per-window?
        nsr = NSIntersectionRect(nsr, bounds);

        // Disallow empty rects
        if (nsr.size.width <= 0  ||  nsr.size.height <= 0) continue;

        while (h--) {
            unsigned long *srcp = srcline;
            unsigned long *dstp = dstline;
            int w = dstWidth;

            if(mInhibitLevel>=10) {	// just update the window bits for later screen draw...
                while(w--) {
                    newval = *srcp;
                    if((mDrawAlpha==0xffffffff || newval & 0xff000000) && firstDraw==0)
                        srcp++;
                    else {
                        newval = newval & 0x00ffffff;
                        if(newval==(mTranslucentBG & 0x00ffffff))
                            *srcp++ = mTranslucentBG ^ 0xff000000;
                        else
                            *srcp++ = newval | (mDrawAlpha ^ 0xff000000);
                    }
                }
            } else if(mDimmed) {
                while(w--) {
                    newval = *srcp;
                    if(newval & 0xff000000 && firstDraw==0) {
                        *dstp++ = newval ^ 0xff000000;
                    } else if(mDrawAlpha==0xffffffff && firstDraw==0) {
                        *dstp++ = (((newval & 0x00fefefe) + (mTranslucentBG & 0x00fefefe)) >> 1)
                                | 0xff000000;
                        *srcp = newval & 0x00ffffff;
                    } //else if(newval==0xff000000)
                      //  *dstp++ = 0;
                    else {
                        newval = newval & 0x00ffffff;
                        if(newval==(mTranslucentBG & 0x00ffffff)) {
                            *dstp++ = mTranslucentBG;
                            *srcp = mTranslucentBG ^ 0xff000000;
                        } else {
                            *dstp++ = (((newval & 0x00fefefe) + (mTranslucentBG & 0x00fefefe)) >> 1)
                                    | mDrawAlpha;
                            *srcp = newval | (mDrawAlpha ^ 0xff000000);
                        }
                    }
                    srcp++;
                }
            } else {
                while(w--) {
                    newval = *srcp;
                    if((mDrawAlpha==0xffffffff || newval & 0xff000000) && firstDraw==0) {
                        *dstp++ = newval ^ 0xff000000;
                    } //else if(newval==0xff000000)
                      //  *dstp++ = 0;
                    else {
                        newval = newval & 0x00ffffff;
                        if(newval==(mTranslucentBG & 0x00ffffff)) {
                            *dstp++ = mTranslucentBG;
                            *srcp = mTranslucentBG ^ 0xff000000;
                        } else {
                            *dstp++ = newval | mDrawAlpha;
                            *srcp = newval | (mDrawAlpha ^ 0xff000000);
                        }
                    }
                    srcp++;
                }
            }
            srcline += srcWidth;
            dstline += dstWidth;
        }

        if(mInhibitLevel==0) {	// don't want to draw to screen if there is some sort of inhibit on
        
            offsetbits = srcBits;
            //bytesPerRow = dstWidth * 4;

            // Flip r to Cocoa-flipped
            nsr.origin.y = bounds.size.height - nsr.origin.y - nsr.size.height;
            destRect = CGRectMake(nsr.origin.x, nsr.origin.y,
                                nsr.size.width, nsr.size.height);

//    	    dataProviderRef = CGDataProviderCreateDirectAccess(offsetbits,
//     	                           destRect.size.height * bytesPerRow, &cb);
            dataProviderRef = CGDataProviderCreateWithData(NULL,offsetbits,
                                    destRect.size.height * bytesPerRow,NULL);

            imageRef = CGImageCreate(destRect.size.width, destRect.size.height,
                                    bitsPerSample, bitsPerPixel, bytesPerRow,
                                    colorSpace, alpha, dataProviderRef, NULL,
                                    0, kCGRenderingIntentDefault);
//                                	 (1, kCGRenderingIntentDefault);

            CGContextDrawImage(destCtx, destRect, imageRef);
            CGImageRelease(imageRef);
            CGDataProviderRelease(dataProviderRef);
        
        }
      }
    }
    
    return 0;
}

- (void)setDim:(BOOL)value update:(BOOL)update
{
    unsigned long mOldAlpha;
    int dummy;
    fakeBoxRec bounds = {0, 0, [self frame].size.width,
                         [self frame].size.height};
                         
    if(mStoredList==NULL) {
        NSLog(@"Ignoring setDim with mStoredList=NULL");
        return;
    }
    
    if(mDimmed!=value) {
        mChangingRectList = YES;
        // check if the main thread is flushing the rect list
        if(update) {
            if(mFlushingRects) {
                sprintf(lastAbortSet,"setDim");
                mAbortRefreshRects = YES;
                [self WaitForFlushingRects:500 warn:200 who:"setDim"];
                mAbortRefreshRects = NO;
            }
            mDimmed = value;
            if(verboseConsole)
                NSLog(@"HackID %d : complete update for dim (%d)",mHackID,mDimmed);
            mStoredCount=0;
            [self SleepingForBitsShape:500 warn:200 who:"setDim"];
            mChangingRectList = NO;
            mOldAlpha = mDrawAlpha;
            mDrawAlpha = 0xffffffff;
            if(mMiniature || mInhibitLevel>=1000)
                dummy = [self copyToScreenViaShape:&bounds count:0];
            else {
                [self lockFocus];
                dummy = [self copyToScreenViaShape:&bounds count:0];
                [[NSGraphicsContext currentContext] flushGraphics];
                [self unlockFocus];
            }
            mDrawAlpha = mOldAlpha;
            shapeDrawing = 0;
            mDrawing = NO;
        } else {
            [self WaitForFlushingRects:500 warn:200 who:"setDim"];
            if(verboseConsole)
                NSLog(@"HackID %d : flushing rects in setDim (%d)",mHackID,value);
            [self flushStored:nil];	// send nil to force flush, despite mChangingRectList being set
            mChangingRectList = NO;
        }
        mDimmed = value;
    }
}

- (BOOL)shadowChangeInhibited
{
    return mInhibitShadowChange;
}

- (void)setInhibitShadowChange:(BOOL)value
{
    mInhibitShadowChange = value;
}

- (BOOL)getMyMiniature
{
    return mMiniature;
}

- (void)setMyMiniature:(BOOL)value
{
    mMiniature = value;
}

- (void)setMyDrawAlpha:(int)alphaval
{
    // Need to make sure all updates are flushed before starting a new alpha
    mChangingRectList = YES;
    // check if the main thread is flushing the rect list
    if(mFlushingRects) {
        [self WaitForFlushingRects:500 warn:200 who:"setMyDrawAlpha"];
    } else if(mStoredCount>0) {
        if(verboseConsole)
            NSLog(@"HackID %d : flushing rects in setMyDrawAlpha (%d)",mHackID,alphaval);
        [self flushStored:nil];	// send nil to force flush, despite mChangingRectList being set
    }
    mDrawAlpha = (unsigned long) alphaval << 24;
//    if(verboseConsole)
//        NSLog(@"HackID %d : set new drawAlpha: %#lx",mHackID,mDrawAlpha);
    mChangingRectList = NO;
}

- (void)setMyTranslucentRed:(int)redval
{
    // Need to make sure all updates are flushed before starting a new translucent colour
    mChangingRectList = YES;
    // check if the main thread is flushing the rect list
    if(mFlushingRects) {
        [self WaitForFlushingRects:500 warn:200 who:"setMyTranslucentRed"];
    } else if(mStoredCount>0) {
        if(verboseConsole)
            NSLog(@"HackID %d : flushing rects in setMyTranslucentRed (%d)",mHackID,redval);
        [self flushStored:nil];	// send nil to force flush, despite mChangingRectList being set
    }
    mTranslucentBG = (mTranslucentBG & 0xffffff00) | (redval & 0xff);
    mChangingRectList = NO;
}

- (void)setMyTranslucentGreen:(int)greenval
{
    // Need to make sure all updates are flushed before starting a new translucent colour
    mChangingRectList = YES;
    // check if the main thread is flushing the rect list
    if(mFlushingRects) {
        [self WaitForFlushingRects:500 warn:200 who:"setMyTranslucentGreen"];
    } else if(mStoredCount>0) {
        if(verboseConsole)
            NSLog(@"HackID %d : flushing rects in setMyTranslucentGreen (%d)",mHackID,greenval);
        [self flushStored:nil];	// send nil to force flush, despite mChangingRectList being set
    }
    mTranslucentBG = (mTranslucentBG & 0xffff00ff) | (greenval << 8);
    mChangingRectList = NO;
}

- (void)setMyTranslucentBlue:(int)blueval
{
    // Need to make sure all updates are flushed before starting a new translucent colour
    mChangingRectList = YES;
    // check if the main thread is flushing the rect list
    if(mFlushingRects) {
        [self WaitForFlushingRects:500 warn:200 who:"setMyTranslucentBlue"];
    } else if(mStoredCount>0) {
        if(verboseConsole)
            NSLog(@"HackID %d : flushing rects in setMyTranslucentBlue (%d)",mHackID,blueval);
        [self flushStored:nil];	// send nil to force flush, despite mChangingRectList being set
    }
    mTranslucentBG = (mTranslucentBG & 0xff00ffff) | ((unsigned long) blueval << 16);
    mChangingRectList = NO;
}

- (void)setMyTranslucentAlpha:(int)alphaval
{
    // Need to make sure all updates are flushed before starting a new translucent alpha
    mChangingRectList = YES;
    // check if the main thread is flushing the rect list
    if(mFlushingRects) {
        [self WaitForFlushingRects:500 warn:200 who:"setMyTranslucentAlpha"];
    } else if(mStoredCount>0) {
        if(verboseConsole)
            NSLog(@"HackID %d : flushing rects in setMyTranslucentAlpha (%d)",mHackID,alphaval);
        [self flushStored:nil];	// send nil to force flush, despite mChangingRectList being set
    }
    mTranslucentBG = (mTranslucentBG & 0x00ffffff) | ((unsigned long) alphaval << 24);
    mChangingRectList = NO;
}

- (void)setMyHackID:(int)hackid
{
    mHackID = hackid;
    mRectTag = [self addTrackingRect:[self frame] owner:self userData:nil assumeInside:NO];
    //NSLog(@"HACKID %d : created tracking rect... %d",mHackID,mRectTag);   
}

- (int)getMyHackID;
{
    return mHackID;
}

- (NSTrackingRectTag)getMyTrackingRect
{
    return mRectTag;
}

- (void)setMyTrackingRect:(NSTrackingRectTag)tag
{
    mRectTag=tag;
}
/*
- (void)setInhibitFull:(BOOL)val
{
    mInhibitUpdate = val;
    NSLog(@"HackID %d : Inhibit update set to %d",mHackID,val);
}

- (void)setInhibitRefresh:(BOOL)val
{
    mInhibitRefresh = val;
    NSLog(@"HackID %d : Inhibit refresh set to %d",mHackID,val);
}
*/
- (void)setInhibitLevel:(BOOL)val
{
    if(mInhibitLevel<1000)
        mInhibitLevel = val;
    else
        NSLog(@"HackID %d is hidden: inhibit level not set to %d",mHackID,val);
}

- (void)setMyHide:(BOOL) val
{
    BOOL wasMin;
    unsigned long mOldAlpha;
    int dummy;
    
    fakeBoxRec bounds = {0, 0, [self frame].size.width,
                         [self frame].size.height};
    
    if(val) {
        NSLog(@"HackID %d: hiding window",mHackID);
        mInhibitLevel = 1000;
        [[self window] orderOut:nil];
    } else if(mInhibitLevel==1000) {
        NSLog(@"HackID %d: showing window",mHackID);
//        wasMin=mMiniature;
        mInhibitLevel = 0;
        if(!mMiniature) {
            [[self window] orderFront:nil];
            if(mFlushingRects) {
                sprintf(lastAbortSet,"setMyHide");
                mAbortRefreshRects = YES;
                [self WaitForFlushingRects:500 warn:200 who:"setMyHide"];
                mAbortRefreshRects = NO;
            }
            //NSLog(@"HackID %d : complete update for setMyHide (%d)",mHackID,mDimmed);
            mStoredCount=0;
            [self SleepingForBitsShape:500 warn:200 who:"setMyHide"];
            mChangingRectList = NO;
            mOldAlpha = mDrawAlpha;
            mDrawAlpha = 0xffffffff;
//            if(mMiniature)
//                [self copyToScreenViaShape:&bounds count:0];
//            else {
                [self lockFocus];
                dummy = [self copyToScreenViaShape:&bounds count:0];
                [[NSGraphicsContext currentContext] flushGraphics];
                [self unlockFocus];
                [self lockFocus];
                [[self window] setHasShadow:NO];
                [[self window] setHasShadow:YES];
                [[NSGraphicsContext currentContext] flushGraphics];
                [self unlockFocus];
//            }
            mDrawAlpha = mOldAlpha;
            shapeDrawing = 0;
            mDrawing = NO;
        }
/*        if(mWaitForDrawRect) {
            int i=0;
            usleep(2000);
            while(mWaitForDrawRect && i<150) {
                i++;
                if(i>400) NSLog(@"Waiting for drawRect in setMyHide");
                usleep(2000);
            }
        }*/
//        mMiniature=wasMin;
    } else {
        NSLog(@"HackID %d: setMyHide(%d) called with inhibit level %d",mHackID,val,mInhibitLevel);
    }
}

- (BOOL)getMyHidden
{
    return (mInhibitLevel==1000);
}

- (BOOL)getWaitingForTimer
{
    return mWaitingForTimer;
}

@end
