/**************************************************************
 *
 * Support for using the Quartz Window Manager cursor
 * NSCursor version
 *
 **************************************************************/
/* $XFree86: xc/programs/Xserver/hw/darwin/bundle/quartzCursor.c,v 1.15 2001/12/22 05:28:35 torrey Exp $ */

#import <AppKit/NSCursor.h>
#import <AppKit/NSImage.h>
#import <AppKit/NSBitmapImageRep.h>
#import <Foundation/NSAutoreleasePool.h>

// hack
#define __EMX__

#include "quartzCommon.h"
#include "quartzCursor.h"
#include "darwin.h"

#include "mi.h"
#include "scrnintstr.h"
#include "cursorstr.h"
#include "mipointrst.h"
#include "globals.h"

// hack
#undef __EMX__

// Maximum NSCursor size
#define CURSORWIDTH 32
#define CURSORHEIGHT 32

// Cursor debugging
#define CURSOR_DEBUG ErrorF

typedef struct {
    int                     qdCursorMode;
    int                     qdCursorVisible;
    int                     useQDCursor;
    QueryBestSizeProcPtr    QueryBestSize;
    miPointerSpriteFuncPtr  spriteFuncs;
} QuartzCursorScreenRec, *QuartzCursorScreenPtr;

static int darwinCursorScreenIndex = -1;
static unsigned long darwinCursorGeneration = 0;
static  CursorPtr quartzLatentCursor = NULL;

#define CURSOR_PRIV(pScreen) \
    ((QuartzCursorScreenPtr)pScreen->devPrivates[darwinCursorScreenIndex].ptr)

#define HIDE_QD_CURSOR(pScreen, visible)                                \
    if (visible) {                                                      \
        int ix;                                                         \
        for (ix = 0; ix < QUARTZ_PRIV(pScreen)->displayCount; ix++) {   \
            CGDisplayHideCursor(QUARTZ_PRIV(pScreen)->displayIDs[ix]);  \
        }                                                               \
        visible = FALSE;                                                \
    } ((void)0)

#define SHOW_QD_CURSOR(pScreen, visible)                                \
    {                                                                   \
        int ix;                                                         \
        for (ix = 0; ix < QUARTZ_PRIV(pScreen)->displayCount; ix++) {   \
            CGDisplayShowCursor(QUARTZ_PRIV(pScreen)->displayIDs[ix]);  \
        }                                                               \
        visible = TRUE;                                                 \
    } ((void)0)




/*
 * MakeNSCursor: construct an NSCursor from an X11 CursorPtr
 * pCursor must be no larger than 16x16.
 *
 * X11 cursor is 1-bit source and 1-bit mask and fore and back colors.
 * The mapping from X11 cursor bits to X11 colors to NSImage colors is:
 * (source,mask | X11 color   | NSImage color)
 *          00  | transparent | transparent
 *          01  | back color  | back color
 *          10  | undefined   | transparent
 *          11  | fore color  | fore color
 */
static NSCursor *MakeNSCursor(CursorPtr pCursor)
{
    int w, h;
    int width = pCursor->bits->width;
    int height = pCursor->bits->height;
    NSPoint hotSpot = NSMakePoint(pCursor->bits->xhot, pCursor->bits->yhot);

    int factor = 1;
    int nswidth = 16; // width and height of NSImage
    int nsheight = 16; // NSCursor requires exactly 16x16 pixels

    // Image bits are 32-bit RGBA
    // X11 color components are 16-bit - shift to get 8-bit
    unsigned int transparent = 0x00000000;
    unsigned int foreColor = ((((pCursor->foreRed   >> 8) & 0xff) << 24) |
                              (((pCursor->foreGreen >> 8) & 0xff) << 16) |
                              (((pCursor->foreBlue  >> 8) & 0xff) << 8)  |
                              0xff);
    unsigned int backColor = ((((pCursor->backRed   >> 8) & 0xff) << 24) |
                              (((pCursor->backGreen >> 8) & 0xff) << 16) |
                              (((pCursor->backBlue  >> 8) & 0xff) << 8)  |
                              0xff);
    
    NSBitmapImageRep *imageRep;
    unsigned int *imageBits;
    NSImage *image;
    NSCursor *nsCursor = nil;
    NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];

    // deal with cursors > 16x16
    if(width>16 || height>16) {
       hotSpot = NSMakePoint(pCursor->bits->xhot/2, pCursor->bits->yhot/2);
       factor = 2;
    }
    
    // Create a cursor bitmap
    imageRep = [[[NSBitmapImageRep alloc] 
                    initWithBitmapDataPlanes:NULL
                    pixelsWide:nswidth    pixelsHigh:nsheight
                    bitsPerSample:8       samplesPerPixel:4
                    hasAlpha:YES          isPlanar:NO
                    colorSpaceName:NSCalibratedRGBColorSpace
                    bytesPerRow:nswidth*4 bitsPerPixel:32] autorelease];
    if (!imageRep) goto done;
    imageBits = (unsigned int *)[imageRep bitmapData];
    bzero(imageBits, nswidth * nsheight * 4);

    if(factor==2) {
    // Draw the cursor in the bitmap.
        CURSOR_DEBUG("Cursor (%d x %d):\n", width, height);
        CURSOR_DEBUG("foreground 0x%x ('#'), background 0x%x ('+')\n", 
                    foreColor, backColor);
            
        for (h = 0; h < height; h++) {
        
            unsigned long sourceRow = 
                (pCursor->bits->source[h*4]   << 24) | (pCursor->bits->source[h*4+1] << 16) |
                (pCursor->bits->source[h*4+2] << 8)  | pCursor->bits->source[h*4+3];
            unsigned long maskRow = 
                (pCursor->bits->mask[h*4]   << 24) | (pCursor->bits->mask[h*4+1] << 16 ) |
                (pCursor->bits->mask[h*4+2] << 8)  | pCursor->bits->mask[h*4+3];
            unsigned long bitmask = 0x80000000;
            
            for (w = 0; w < width; w++, bitmask >>= 1) {
                if ((maskRow & bitmask) == 0) CURSOR_DEBUG(" ");
                else if ((sourceRow & bitmask) == 0) CURSOR_DEBUG("+");
                else CURSOR_DEBUG("#");
            }
            CURSOR_DEBUG("|\n");
        }
    }
    
    for (h = 0; h < height; h+=factor) {
        // 16-bit row. High width bits are the bits to use.
        unsigned long sourceRow = 
            (pCursor->bits->source[h*4] << 8) | pCursor->bits->source[h*4+1];
        unsigned long maskRow = 
            (pCursor->bits->mask[h*4] << 8) | pCursor->bits->mask[h*4+1];
        unsigned long bitmask = 0x8000;

        if(factor==2) {

            unsigned long sourceRow2 = 
                (pCursor->bits->source[h*4+4] << 24) | (pCursor->bits->source[h*4+5] << 16) |
                (pCursor->bits->source[h*4+6] << 8)  | pCursor->bits->source[h*4+7];
            unsigned long maskRow2 = 
                (pCursor->bits->mask[h*4+4] << 24) | (pCursor->bits->mask[h*4+5] << 16 ) |
                (pCursor->bits->mask[h*4+6] << 8)  | pCursor->bits->mask[h*4+7];
        
            sourceRow = (sourceRow << 16) | (pCursor->bits->source[h*4+2] << 8) | pCursor->bits->source[h*4+3];
            maskRow   = (maskRow << 16)   | (pCursor->bits->mask[h*4+2] << 8)   | pCursor->bits->mask[h*4+3];
            bitmask   = bitmask << 16;
            
            for (w = 0; w < width; w+=2, bitmask >>= 2) {
                unsigned int color;
                unsigned long bitmask2 = bitmask >> 1;
                
//                if((maskRow & bitmask)  == 0 && (maskRow2 & bitmask)  == 0 &
//                   (maskRow & bitmask2) == 0 && (maskRow2 & bitmask2) == 0)
//                   color = transparent;
//                else {
                    if((maskRow & bitmask) == 0)
                        color = transparent;
                    else if((sourceRow & bitmask) == 0)
                        color = (backColor & 0xFCFCFCFC) >> 2;
                    else
                        color = (foreColor & 0xFCFCFCFC) >> 2;
                        
                    if((maskRow2 & bitmask) == 0)
                        ;
                    else if((sourceRow2 & bitmask) == 0)
                        color = color + ((backColor & 0xFCFCFCFC) >> 2);
                    else
                        color = color + ((foreColor & 0xFCFCFCFC) >> 2);
                        
                    if((maskRow & bitmask2) == 0)
                        ;
                    else if((sourceRow & bitmask2) == 0)
                        color = color + ((backColor & 0xFCFCFCFC) >> 2);
                    else
                        color = color + ((foreColor & 0xFCFCFCFC) >> 2);
                        
                    if((maskRow2 & bitmask2) == 0)
                        ;
                    else if((sourceRow2 & bitmask2) == 0)
                        color = color + ((backColor & 0xFCFCFCFC) >> 2);
                    else
                        color = color + ((foreColor & 0xFCFCFCFC) >> 2);

                    if((color & 0xFF) > 0) color |= 0xFF;
                    if((color & 0xFF00) > 0xFB00) color |= 0xFF00;
                    if((color & 0xFF0000) > 0xFB0000) color |= 0xFF0000;
                    if((color & 0xFF000000) > 0xFB000000) color |= 0xFF000000;
//                }
            
                imageBits[(h*nswidth + w)/2] = color;

                if (color == transparent) CURSOR_DEBUG(" ");
                else if (color == backColor) CURSOR_DEBUG("+");
                else if (color == foreColor) CURSOR_DEBUG("#");
                else CURSOR_DEBUG("x");
            }
            CURSOR_DEBUG("|\n");

        } else {
        
            for (w = 0; w < width; w++, bitmask >>= 1) {
                unsigned int color;
                if ((maskRow & bitmask) == 0) color = transparent;
                else if ((sourceRow & bitmask) == 0) color = backColor;
                else color = foreColor;
            
                imageBits[h*nswidth + w] = color;

//                if (color == transparent) CURSOR_DEBUG(" ");
//                else if (color == backColor) CURSOR_DEBUG("+");
//                else CURSOR_DEBUG("#");
            }
//            CURSOR_DEBUG("|\n");

        }
    }
    if(verboseConsole) ErrorF("XDarwin info: created a new NSCursor\n");
    // Create an NSImage from the cursor bitmap.
    image = [[[NSImage alloc] initWithSize:[imageRep size]] autorelease];
    if (!image) goto done;
    [image addRepresentation:imageRep];

    // Create an NSCursor from the image.
    nsCursor = [[NSCursor alloc] initWithImage:image hotSpot:hotSpot];

    done:
    [pool release];

    return nsCursor;
}


/*
===========================================================================

 Pointer sprite functions

===========================================================================
*/


/*
 * A realized cursor's devPrivate is always one of the following:
 * - a constructed NSCursor 
 * - a retained copy of [NSCursor arrowCursor]
 * - nil (iff using X11 software cursor in full-screen mode)
 * In particular:
 * - a nil devPrivate always means X11 software cursor is in use
 * - a non-nil devPrivate is always a retained NSCursor.
 */

/*
 * QuartzRealizeCursor
 * Convert the X cursor representation to QuickDraw format if possible.
 */
Bool
QuartzRealizeCursor(
    ScreenPtr pScreen,
    CursorPtr pCursor )
{
    Bool result;
    NSCursor *nsCursor;
    QuartzCursorScreenPtr ScreenPriv = CURSOR_PRIV(pScreen);

    if(!pCursor || !pCursor->bits)
        return FALSE;
 
    if (pCursor->bits->height <= CURSORHEIGHT  &&  
        pCursor->bits->width  <= CURSORWIDTH  &&  
        ScreenPriv->useQDCursor) 
    {
        // Cursor is small enough for NSCursor
        nsCursor = MakeNSCursor(pCursor);
        result = (nsCursor != nil);
    } else {
        // Cursor is too big for NSCursor. 
        if (quartzRootless) {
            // Rootless mode can't use a software cursor - use an arrow instead
            nsCursor = [[NSCursor arrowCursor] retain];
            result = TRUE;
        } else {
            // Full-screen mode can use an X11 software cursor
            nsCursor = nil;
            result = (*ScreenPriv->spriteFuncs->RealizeCursor)
                (pScreen, pCursor);
        }
    }

    // save the NSCursor, if any
    pCursor->devPriv[pScreen->myNum] = (pointer) nsCursor;

    return result;
}


/*
 * QuartzUnrealizeCursor
 * Free the storage space associated with a realized cursor.
 */
Bool
QuartzUnrealizeCursor(
    ScreenPtr pScreen,
    CursorPtr pCursor )
{
    Bool result;
    QuartzCursorScreenPtr ScreenPriv = CURSOR_PRIV(pScreen);
    NSCursor *nsCursor = (NSCursor *)pCursor->devPriv[pScreen->myNum];

    if (nsCursor) {
        // We constructed an NSCursor or retained a copy of the arrow. 
        [nsCursor release];
        result = TRUE;
    } else if (quartzRootless) {
        // This should never happen.
        ErrorF("QuartzUnrealizeCursor - nil cursor in rootless mode!\n");
        result = FALSE;
    } else {
        // We used a software cursor
        result = (*ScreenPriv->spriteFuncs->UnrealizeCursor)
            (pScreen, pCursor);
    }

    pCursor->devPriv[pScreen->myNum] = NULL;

    return result;
}


/*
 * QuartzSetCursor
 * Set the cursor sprite and position.
 * Use QuickDraw cursor if possible.
 */
static void
QuartzSetCursor(
    ScreenPtr       pScreen,
    CursorPtr       pCursor,
    int             x,
    int             y)
{
    QuartzCursorScreenPtr ScreenPriv = CURSOR_PRIV(pScreen);
    NSCursor *nsCursor = pCursor ? (NSCursor *)pCursor->devPriv[pScreen->myNum]:nil;

    if(verboseConsole) ErrorF("XDarwin info: into QuartzSetCursor (%#lx)\n",pCursor);
    quartzLatentCursor = pCursor;

    // Don't touch Mac OS cursor if X is hidden!
    if (!quartzServerVisible)
        return;

    if (!pCursor) {
        if(verboseConsole) ErrorF("XDarwin info: removing cursor (pCursor=nil)\n");
        // Remove the cursor completely.
        HIDE_QD_CURSOR(pScreen, ScreenPriv->qdCursorVisible);
        if (! ScreenPriv->qdCursorMode)
            (*ScreenPriv->spriteFuncs->SetCursor)(pScreen, 0, x, y);
    }
    else if (nsCursor != nil) {
        // An NSCursor is available. Use it.
        if (! ScreenPriv->qdCursorMode) {
            // remove the X cursor
            (*ScreenPriv->spriteFuncs->SetCursor)(pScreen, 0, x, y);
            ScreenPriv->qdCursorMode = TRUE;
        }
        if(verboseConsole) ErrorF("XDarwin info: changing NSCursor\n");
        [nsCursor set];
        SHOW_QD_CURSOR(pScreen, ScreenPriv->qdCursorVisible);
    }
    else if (quartzRootless) {
        // This should never happen.
        ErrorF("QuartzSetCursor: nil cursor in rootless mode!");
    }
    else {
        // No NSCursor available. Use X11 software cursor instead.
        HIDE_QD_CURSOR(pScreen, ScreenPriv->qdCursorVisible);
        ScreenPriv->qdCursorMode = FALSE;
        (*ScreenPriv->spriteFuncs->SetCursor)(pScreen, pCursor, x, y);
    }
}


/*
 * QuartzMoveCursor
 * Move the cursor. This is a noop for QuickDraw.
 */
static void
QuartzMoveCursor(
    ScreenPtr   pScreen,
    int         x,
    int         y)
{
    QuartzCursorScreenPtr ScreenPriv = CURSOR_PRIV(pScreen);

    // only the X11 software cursor needs to be explicitly moved
    if (!ScreenPriv->qdCursorMode)
        (*ScreenPriv->spriteFuncs->MoveCursor)(pScreen, x, y);
}


static miPointerSpriteFuncRec quartzSpriteFuncsRec = {
    QuartzRealizeCursor,
    QuartzUnrealizeCursor,
    QuartzSetCursor,
    QuartzMoveCursor
};

/*
===========================================================================

 Pointer screen functions

===========================================================================
*/

/*
 * QuartzCursorOffScreen
 */
static Bool QuartzCursorOffScreen(ScreenPtr *pScreen, int *x, int *y)
{
    return FALSE;
}


/*
 * QuartzCrossScreen
 */
static void QuartzCrossScreen(ScreenPtr pScreen, Bool entering)
{
    return;
}


/*
 * QuartzWarpCursor
 *  Change the cursor position without generating an event or motion history.
 *  The input coordinates (x,y) are in pScreen-local X11 coordinates.
 *
 */
static void
QuartzWarpCursor(
    ScreenPtr               pScreen,
    int                     x,
    int                     y)
{
    static int              neverMoved = TRUE;

    if (neverMoved) {
        // Don't move the cursor the first time. This is the jump-to-center
        // initialization, and it's annoying because we may still be in MacOS.
        neverMoved = FALSE;
        return;
    }

    if (quartzServerVisible) {
        CGDisplayErr        cgErr;
        CGPoint             cgPoint;
        // Only need to do this for one display. Any display will do.
        CGDirectDisplayID   cgID = QUARTZ_PRIV(pScreen)->displayIDs[0];
        CGRect              cgRect = CGDisplayBounds(cgID);

        // Convert (x,y) to CoreGraphics screen-local CG coordinates.
        // This is necessary because the X11 screen and CG screen may not
        // coincide. (e.g. X11 screen may be moved to dodge the menu bar)

        // Make point in X11 global coordinates
        cgPoint = CGPointMake(x + dixScreenOrigins[pScreen->myNum].x,
                              y + dixScreenOrigins[pScreen->myNum].y);
        // Shift to CoreGraphics global screen coordinates
        cgPoint.x += darwinMainScreenX;
        cgPoint.y += darwinMainScreenY;
        // Shift to CoreGraphics screen-local coordinates
        cgPoint.x -= cgRect.origin.x;
        cgPoint.y -= cgRect.origin.y;

        cgErr = CGDisplayMoveCursorToPoint(cgID, cgPoint);
        if (cgErr != CGDisplayNoErr) {
            ErrorF("Could not set cursor position with error code 0x%x.\n",
                    cgErr);
        }
    }

    miPointerWarpCursor(pScreen, x, y);
}


static miPointerScreenFuncRec quartzScreenFuncsRec = {
    QuartzCursorOffScreen,
    QuartzCrossScreen,
    QuartzWarpCursor,
};

/*
===========================================================================

 Other screen functions

===========================================================================
*/

/*
 * QuartzCursorQueryBestSize
 * Handle queries for best cursor size
 */
static void
QuartzCursorQueryBestSize(
   int              class,
   unsigned short   *width,
   unsigned short   *height,
   ScreenPtr        pScreen)
{
    QuartzCursorScreenPtr ScreenPriv = CURSOR_PRIV(pScreen);

    if (class == CursorShape) {
        *width = CURSORWIDTH;
        *height = CURSORHEIGHT;
    } else {
        (*ScreenPriv->QueryBestSize)(class, width, height, pScreen);
    }
}


/*
 * QuartzInitCursor
 * Initialize cursor support
 */
Bool
QuartzInitCursor(
    ScreenPtr   pScreen )
{
    QuartzCursorScreenPtr   ScreenPriv;
    miPointerScreenPtr      PointPriv;
    DarwinFramebufferPtr    dfb = SCREEN_PRIV(pScreen);

    // initialize software cursor handling (always needed as backup)
    if (!miDCInitialize(pScreen, &quartzScreenFuncsRec)) {
        return FALSE;
    }

    // allocate private storage for this screen's QuickDraw cursor info
    if (darwinCursorGeneration != serverGeneration) {
        if ((darwinCursorScreenIndex = AllocateScreenPrivateIndex()) < 0)
            return FALSE;
        darwinCursorGeneration = serverGeneration;
    }

    ScreenPriv = xcalloc( 1, sizeof(QuartzCursorScreenRec) );
    if (!ScreenPriv) return FALSE;

    CURSOR_PRIV(pScreen) = ScreenPriv;

    // override some screen procedures
    ScreenPriv->QueryBestSize = pScreen->QueryBestSize;
    pScreen->QueryBestSize = QuartzCursorQueryBestSize;

    // initialize QuickDraw cursor handling
    PointPriv = (miPointerScreenPtr)
                    pScreen->devPrivates[miPointerScreenIndex].ptr;

    ScreenPriv->spriteFuncs = PointPriv->spriteFuncs;
    PointPriv->spriteFuncs = &quartzSpriteFuncsRec;

    if (!quartzRootless)
        ScreenPriv->useQDCursor = QuartzFSUseQDCursor(dfb->colorBitsPerPixel);
    else
        ScreenPriv->useQDCursor = TRUE;
    ScreenPriv->qdCursorMode = TRUE;
    ScreenPriv->qdCursorVisible = TRUE;
    return TRUE;
}


// X server is hiding. Restore the Aqua cursor.
void QuartzSuspendXCursor(
    ScreenPtr pScreen )
{
    QuartzCursorScreenPtr ScreenPriv = CURSOR_PRIV(pScreen);

    [[NSCursor arrowCursor] set];
    SHOW_QD_CURSOR(pScreen, ScreenPriv->qdCursorVisible);
}


// X server is showing. Restore the X cursor.
void QuartzResumeXCursor(
    ScreenPtr pScreen,
    int x,
    int y )
{
    QuartzSetCursor(pScreen, quartzLatentCursor, x, y);
}
